"""
Definitions for spherical coordinate systems




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------
from __future__ import print_function
import numpy as np
from .coordinate_handler import \
    CoordinateHandler, \
    _unknown_coord, \
    _get_coord_fields
from yt.utilities.lib.pixelization_routines import \
    pixelize_cylinder, pixelize_aitoff

class SphericalCoordinateHandler(CoordinateHandler):
    name = "spherical"

    def __init__(self, ds, ordering = ('r', 'theta', 'phi')):
        super(SphericalCoordinateHandler, self).__init__(ds, ordering)
        # Generate 
        self.image_units = {}
        self.image_units[self.axis_id['r']] = ("rad", "rad")
        self.image_units[self.axis_id['theta']] = (None, None)
        self.image_units[self.axis_id['phi']] = (None, None)

    def setup_fields(self, registry):
        # return the fields for r, z, theta
        registry.add_field(("index", "dx"), sampling_type="cell",  function=_unknown_coord)
        registry.add_field(("index", "dy"), sampling_type="cell",  function=_unknown_coord)
        registry.add_field(("index", "dz"), sampling_type="cell",  function=_unknown_coord)
        registry.add_field(("index", "x"), sampling_type="cell",  function=_unknown_coord)
        registry.add_field(("index", "y"), sampling_type="cell",  function=_unknown_coord)
        registry.add_field(("index", "z"), sampling_type="cell",  function=_unknown_coord)
        f1, f2 = _get_coord_fields(self.axis_id['r'])
        registry.add_field(("index", "dr"), sampling_type="cell",  function = f1,
                           display_field = False,
                           units = "code_length")
        registry.add_field(("index", "r"), sampling_type="cell",  function = f2,
                           display_field = False,
                           units = "code_length")

        f1, f2 = _get_coord_fields(self.axis_id['theta'], "")
        registry.add_field(("index", "dtheta"), sampling_type="cell",  function = f1,
                           display_field = False,
                           units = "")
        registry.add_field(("index", "theta"), sampling_type="cell",  function = f2,
                           display_field = False,
                           units = "")

        f1, f2 = _get_coord_fields(self.axis_id['phi'], "")
        registry.add_field(("index", "dphi"), sampling_type="cell",  function = f1,
                           display_field = False,
                           units = "")
        registry.add_field(("index", "phi"), sampling_type="cell",  function = f2,
                           display_field = False,
                           units = "")

        def _SphericalVolume(field, data):
            # Here we compute the spherical volume element exactly
            r = data["index", "r"]
            dr = data["index", "dr"]
            theta = data["index", "theta"]
            dtheta = data["index", "dtheta"]
            vol = ((r+0.5*dr)**3-(r-0.5*dr)**3)/3.0
            vol *= np.cos(theta-0.5*dtheta)-np.cos(theta+0.5*dtheta)
            vol *= data["index", "dphi"]
            return vol
        registry.add_field(("index", "cell_volume"), sampling_type="cell", 
                 function=_SphericalVolume,
                 units = "code_length**3")

        def _path_r(field, data):
            return data["index", "dr"]
        registry.add_field(("index", "path_element_r"), sampling_type="cell", 
                 function = _path_r,
                 units = "code_length")
        def _path_theta(field, data):
            # Note: this already assumes cell-centered
            return data["index", "r"] * data["index", "dtheta"]
        registry.add_field(("index", "path_element_theta"), sampling_type="cell", 
                 function = _path_theta,
                 units = "code_length")
        def _path_phi(field, data):
            # Note: this already assumes cell-centered
            return data["index", "r"] \
                    * data["index", "dphi"] \
                    * np.sin(data["index", "theta"])
        registry.add_field(("index", "path_element_phi"), sampling_type="cell", 
                 function = _path_phi,
                 units = "code_length")

    def pixelize(self, dimension, data_source, field, bounds, size,
                 antialias = True, periodic = True):
        self.period
        name = self.axis_name[dimension]
        if name == 'r':
            return self._ortho_pixelize(data_source, field, bounds, size,
                                        antialias, dimension, periodic)
        elif name in ('theta', 'phi'):
            return self._cyl_pixelize(data_source, field, bounds, size,
                                          antialias, dimension)
        else:
            raise NotImplementedError

    def _ortho_pixelize(self, data_source, field, bounds, size, antialias,
                        dim, periodic):
        buff = pixelize_aitoff(data_source["py"], data_source["pdy"],
                               data_source["px"], data_source["pdx"],
                               size, data_source[field], None,
                               None, theta_offset = 0,
                               phi_offset = 0).transpose()
        return buff


    def _cyl_pixelize(self, data_source, field, bounds, size, antialias,
                      dimension):
        name = self.axis_name[dimension]
        buff = np.zeros((size[1], size[0]), dtype="f8")
        if name == 'theta':
            pixelize_cylinder(buff,
                              data_source['px'],
                              data_source['pdx'],
                              data_source['py'],
                              data_source['pdy'],
                              data_source[field], bounds)
        elif name == 'phi':
            # Note that we feed in buff.T here
            pixelize_cylinder(buff.T,
                             data_source['px'],
                             data_source['pdx'],
                             data_source['py'],
                             data_source['pdy'],
                             data_source[field], bounds)
        else:
            raise RuntimeError
        return buff


    def convert_from_cartesian(self, coord):
        raise NotImplementedError

    def convert_to_cartesian(self, coord):
        if isinstance(coord, np.ndarray) and len(coord.shape) > 1:
            ri = self.axis_id['r']
            thetai = self.axis_id['theta']
            phii = self.axis_id['phi']
            r = coord[:,ri]
            theta = coord[:,thetai]
            phi = coord[:,phii]
            nc = np.zeros_like(coord)
            # r, theta, phi
            nc[:,ri] = np.cos(phi) * np.sin(theta)*r
            nc[:,thetai] = np.sin(phi) * np.sin(theta)*r
            nc[:,phii] = np.cos(theta) * r
        else:
            r, theta, phi = coord
            nc = (np.cos(phi) * np.sin(theta)*r,
                  np.sin(phi) * np.sin(theta)*r,
                  np.cos(theta) * r)
        return nc

    def convert_to_cylindrical(self, coord):
        raise NotImplementedError

    def convert_from_cylindrical(self, coord):
        raise NotImplementedError

    def convert_to_spherical(self, coord):
        raise NotImplementedError

    def convert_from_spherical(self, coord):
        raise NotImplementedError

    _image_axis_name = None
    @property
    def image_axis_name(self):    
        if self._image_axis_name is not None:
            return self._image_axis_name
        # This is the x and y axes labels that get displayed.  For
        # non-Cartesian coordinates, we usually want to override these for
        # Cartesian coordinates, since we transform them.
        rv = {self.axis_id['r']: ('theta', 'phi'),
              self.axis_id['theta']: ('x / \\sin(\\theta)', 'y / \\sin(\\theta)'),
              self.axis_id['phi']: ('R', 'z')}
        for i in list(rv.keys()):
            rv[self.axis_name[i]] = rv[i]
            rv[self.axis_name[i].capitalize()] = rv[i]
        self._image_axis_name = rv
        return rv

    _x_pairs = (('r', 'theta'), ('theta', 'r'), ('phi', 'r'))
    _y_pairs = (('r', 'phi'), ('theta', 'phi'), ('phi', 'theta'))

    @property
    def period(self):
        return self.ds.domain_width

    def sanitize_center(self, center, axis):
        center, display_center = super(
            SphericalCoordinateHandler, self).sanitize_center(center, axis)
        name = self.axis_name[axis]
        if name == 'r':
            display_center = center
        elif name == 'theta':
            display_center = (0.0 * display_center[0],
                              0.0 * display_center[1],
                              0.0 * display_center[2])
        elif name == 'phi':
            display_center = [self.ds.domain_width[0]/2.0 +
                                self.ds.domain_left_edge[0],
                              0.0 * display_center[1],
                              0.0 * display_center[2]]
            ri = self.axis_id['r']
            c = self.ds.domain_width[ri]/2.0 + self.ds.domain_left_edge[ri]
            display_center[ri] = c
            display_center = tuple(display_center)
        return center, display_center

    def sanitize_width(self, axis, width, depth):
        name = self.axis_name[axis]
        if width is not None:
            width = super(SphericalCoordinateHandler, self).sanitize_width(
              axis, width, depth)
        elif name == 'r':
            width = [self.ds.domain_width[self.x_axis['r']],
                     self.ds.domain_width[self.y_axis['r']]]
        elif name == 'theta':
            ri = self.axis_id['r']
            # Remember, in spherical coordinates when we cut in theta,
            # we create a conic section
            width = [2.0*self.ds.domain_width[ri],
                     2.0*self.ds.domain_width[ri]]
        elif name == 'phi':
            ri = self.axis_id['r']
            width = [self.ds.domain_right_edge[ri],
                     2.0*self.ds.domain_width[ri]]
        return width

    def _sanity_check(self):
        """This prints out a handful of diagnostics that help verify the
        dataset is well formed."""
        # We just check a few things here.
        dd = self.ds.all_data()
        r0 = self.ds.domain_left_edge[self.axis_id['r']]
        r1 = self.ds.domain_right_edge[self.axis_id['r']]
        v1 = 4.0 * np.pi / 3.0 * (r1**3 - r0**3)
        print("Total volume should be 4*pi*r**3 = %0.16e" % (v1))
        v2 = dd.quantities.total_quantity("cell_volume")
        print("Actual volume is                   %0.16e" % (v2))
        print("Relative difference: %0.16e" % (np.abs(v2-v1)/(v2+v1)))




import numpy as np
cimport numpy as np
cimport cython

cdef extern from "platform_dep.h":
    double erf(double x)
    
@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def broaden_lines(np.ndarray[np.float64_t, ndim=1] E0,
                  np.ndarray[np.float64_t, ndim=1] sigma,
                  np.ndarray[np.float64_t, ndim=1] amp,
                  np.ndarray[np.float64_t, ndim=1] ebins):

    cdef int i, j, n, m
    cdef double x, isigma
    cdef np.ndarray[np.float64_t, ndim=1] cdf, vec

    n = E0.shape[0]
    m = ebins.shape[0]
    cdf = np.zeros(m)
    vec = np.zeros(m-1)
    
    for i in range(n):
        isigma = 1.0/sigma[i]
        for j in range(m):
            x = (ebins[j]-E0[i])*isigma
            cdf[j] = 0.5*(1+erf(x))
        for j in range(m-1):
            vec[j] = vec[j] + (cdf[j+1] - cdf[j])*amp[i]
    return vec



"""
Definitions for the lens code




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
cimport numpy as np
cimport cython
from .volume_container cimport VolumeContainer
from vec3_ops cimport dot, subtract, L2_norm, fma
from libc.math cimport exp, floor, log2, \
    fabs, atan, atan2, asin, cos, sin, sqrt, acos, M_PI
from yt.utilities.lib.fp_utils cimport imax, fmax, imin, fmin, iclip, fclip, i64clip
from .image_samplers cimport \
    ImageSampler, \
    calculate_extent_function, \
    generate_vector_info_function

cdef extern from "platform_dep.h":
    long int lrint(double x) nogil

cdef extern from "limits.h":
    cdef int SHRT_MAX

cdef generate_vector_info_function generate_vector_info_plane_parallel
cdef generate_vector_info_function generate_vector_info_null
cdef calculate_extent_function calculate_extent_plane_parallel
cdef calculate_extent_function calculate_extent_perspective
cdef calculate_extent_function calculate_extent_null



"""
API for light_ray



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.funcs import issue_deprecation_warning

issue_deprecation_warning(
    "Development of the LightRay module has been moved to the Trident "
    "package. This version is deprecated and will be removed from yt "
    "in a future release. See https://github.com/trident-project/trident "
    "for further information.")

from .light_ray import \
    LightRay



"""
GadgetSimulation class and member functions.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013-2015, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
import glob
import os

from yt.convenience import \
    load
from yt.funcs import \
    only_on_root
from yt.data_objects.time_series import \
    SimulationTimeSeries, DatasetSeries
from yt.units import dimensions
from yt.units.unit_registry import \
    UnitRegistry
from yt.units.yt_array import \
    YTArray
from yt.utilities.cosmology import \
    Cosmology
from yt.utilities.exceptions import \
    InvalidSimulationTimeSeries, \
    MissingParameter, \
    NoStoppingCondition
from yt.utilities.exceptions import \
    YTOutputNotIdentified
from yt.utilities.logger import ytLogger as \
    mylog
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    parallel_objects

class GadgetSimulation(SimulationTimeSeries):
    r"""
    Initialize an Gadget Simulation object.

    Upon creation, the parameter file is parsed and the time and redshift
    are calculated and stored in all_outputs.  A time units dictionary is
    instantiated to allow for time outputs to be requested with physical
    time units.  The get_time_series can be used to generate a
    DatasetSeries object.

    parameter_filename : str
        The simulation parameter file.
    find_outputs : bool
        If True, the OutputDir directory is searched for datasets.  
        Time and redshift information are gathered by temporarily 
        instantiating each dataset.  This can be used when simulation 
        data was created in a non-standard way, making it difficult 
        to guess the corresponding time and redshift information.
        Default: False.

    Examples
    --------
    >>> import yt
    >>> gs = yt.simulation("my_simulation.par", "Gadget")
    >>> gs.get_time_series()
    >>> for ds in gs:
    ...     print ds.current_time

    """

    def __init__(self, parameter_filename, find_outputs=False):
        self.simulation_type = "particle"
        self.dimensionality = 3
        SimulationTimeSeries.__init__(self, parameter_filename,
                                      find_outputs=find_outputs)

    def _set_units(self):
        self.unit_registry = UnitRegistry()
        self.time_unit = self.quan(1.0, "s")
        if self.cosmological_simulation:
            # Instantiate Cosmology object for units and time conversions.
            self.cosmology = \
              Cosmology(hubble_constant=self.hubble_constant,
                        omega_matter=self.omega_matter,
                        omega_lambda=self.omega_lambda,
                        unit_registry=self.unit_registry)
            self.unit_registry.modify("h", self.hubble_constant)
            # Comoving lengths
            for my_unit in ["m", "pc", "AU", "au"]:
                new_unit = "%scm" % my_unit
                # technically not true, but should be ok
                self.unit_registry.add(
                    new_unit, self.unit_registry.lut[my_unit][0],
                    dimensions.length, "\\rm{%s}/(1+z)" % my_unit)
            self.length_unit = self.quan(self.unit_base["UnitLength_in_cm"],
                                         "cmcm / h", registry=self.unit_registry)
            self.mass_unit = self.quan(self.unit_base["UnitMass_in_g"],
                                         "g / h", registry=self.unit_registry)
            self.box_size = self.box_size * self.length_unit
            self.domain_left_edge = self.domain_left_edge * self.length_unit
            self.domain_right_edge = self.domain_right_edge * self.length_unit
            self.unit_registry.add("unitary", float(self.box_size.in_base()),
                                   self.length_unit.units.dimensions)
        else:
            # Read time from file for non-cosmological sim
            self.time_unit = self.quan(
                self.unit_base["UnitLength_in_cm"]/ \
                    self.unit_base["UnitVelocity_in_cm_per_s"], "s")
            self.unit_registry.add("code_time", 1.0, dimensions.time)
            self.unit_registry.modify("code_time", self.time_unit)
            # Length
            self.length_unit = self.quan(
                self.unit_base["UnitLength_in_cm"],"cm")
            self.unit_registry.add("code_length", 1.0, dimensions.length)
            self.unit_registry.modify("code_length", self.length_unit)

    def get_time_series(self, initial_time=None, final_time=None,
                        initial_redshift=None, final_redshift=None,
                        times=None, redshifts=None, tolerance=None,
                        parallel=True, setup_function=None):

        """
        Instantiate a DatasetSeries object for a set of outputs.

        If no additional keywords given, a DatasetSeries object will be
        created with all potential datasets created by the simulation.

        Outputs can be gather by specifying a time or redshift range
        (or combination of time and redshift), with a specific list of
        times or redshifts), or by simply searching all subdirectories 
        within the simulation directory.

        initial_time : tuple of type (float, str)
            The earliest time for outputs to be included.  This should be 
            given as the value and the string representation of the units.
            For example, (5.0, "Gyr").  If None, the initial time of the 
            simulation is used.  This can be used in combination with 
            either final_time or final_redshift.
            Default: None.
        final_time : tuple of type (float, str)
            The latest time for outputs to be included.  This should be 
            given as the value and the string representation of the units.
            For example, (13.7, "Gyr"). If None, the final time of the 
            simulation is used.  This can be used in combination with either 
            initial_time or initial_redshift.
            Default: None.
        times : tuple of type (float array, str)
            A list of times for which outputs will be found and the units 
            of those values.  For example, ([0, 1, 2, 3], "s").
            Default: None.
        initial_redshift : float
            The earliest redshift for outputs to be included.  If None,
            the initial redshift of the simulation is used.  This can be
            used in combination with either final_time or
            final_redshift.
            Default: None.
        final_redshift : float
            The latest redshift for outputs to be included.  If None,
            the final redshift of the simulation is used.  This can be
            used in combination with either initial_time or
            initial_redshift.
            Default: None.
        redshifts : array_like
            A list of redshifts for which outputs will be found.
            Default: None.
        tolerance : float
            Used in combination with "times" or "redshifts" keywords,
            this is the tolerance within which outputs are accepted
            given the requested times or redshifts.  If None, the
            nearest output is always taken.
            Default: None.
        parallel : bool/int
            If True, the generated DatasetSeries will divide the work
            such that a single processor works on each dataset.  If an
            integer is supplied, the work will be divided into that
            number of jobs.
            Default: True.
        setup_function : callable, accepts a ds
            This function will be called whenever a dataset is loaded.

        Examples
        --------

        >>> import yt
        >>> gs = yt.simulation("my_simulation.par", "Gadget")
        
        >>> gs.get_time_series(initial_redshift=10, final_time=(13.7, "Gyr"))

        >>> gs.get_time_series(redshifts=[3, 2, 1, 0])

        >>> # after calling get_time_series
        >>> for ds in gs.piter():
        ...     p = ProjectionPlot(ds, "x", "density")
        ...     p.save()

        >>> # An example using the setup_function keyword
        >>> def print_time(ds):
        ...     print ds.current_time
        >>> gs.get_time_series(setup_function=print_time)
        >>> for ds in gs:
        ...     SlicePlot(ds, "x", "Density").save()

        """

        if (initial_redshift is not None or \
            final_redshift is not None) and \
            not self.cosmological_simulation:
            raise InvalidSimulationTimeSeries(
                "An initial or final redshift has been given for a " +
                "noncosmological simulation.")

        my_all_outputs = self.all_outputs
        if not my_all_outputs:
            DatasetSeries.__init__(self, outputs=[], parallel=parallel,
                                   unit_base=self.unit_base)
            mylog.info("0 outputs loaded into time series.")
            return

        # Apply selection criteria to the set.
        if times is not None:
            my_outputs = self._get_outputs_by_key("time", times,
                                                  tolerance=tolerance,
                                                  outputs=my_all_outputs)

        elif redshifts is not None:
            my_outputs = self._get_outputs_by_key("redshift",
                                                  redshifts, tolerance=tolerance,
                                                  outputs=my_all_outputs)

        else:
            if initial_time is not None:
                if isinstance(initial_time, float):
                    initial_time = self.quan(initial_time, "code_time")
                elif isinstance(initial_time, tuple) and len(initial_time) == 2:
                    initial_time = self.quan(*initial_time)
                elif not isinstance(initial_time, YTArray):
                    raise RuntimeError(
                        "Error: initial_time must be given as a float or " +
                        "tuple of (value, units).")
            elif initial_redshift is not None:
                my_initial_time = self.cosmology.t_from_z(initial_redshift)
            else:
                my_initial_time = self.initial_time

            if final_time is not None:
                if isinstance(final_time, float):
                    final_time = self.quan(final_time, "code_time")
                elif isinstance(final_time, tuple) and len(final_time) == 2:
                    final_time = self.quan(*final_time)
                elif not isinstance(final_time, YTArray):
                    raise RuntimeError(
                        "Error: final_time must be given as a float or " +
                        "tuple of (value, units).")
                my_final_time = final_time.in_units("s")
            elif final_redshift is not None:
                my_final_time = self.cosmology.t_from_z(final_redshift)
            else:
                my_final_time = self.final_time

            my_initial_time.convert_to_units("s")
            my_final_time.convert_to_units("s")
            my_times = np.array([a["time"] for a in my_all_outputs])
            my_indices = np.digitize([my_initial_time, my_final_time], my_times)
            if my_initial_time == my_times[my_indices[0] - 1]: my_indices[0] -= 1
            my_outputs = my_all_outputs[my_indices[0]:my_indices[1]]

        init_outputs = []
        for output in my_outputs:
            if os.path.exists(output["filename"]):
                init_outputs.append(output["filename"])
        if len(init_outputs) == 0 and len(my_outputs) > 0:
            mylog.warn("Could not find any datasets.  " +
                       "Check the value of OutputDir in your parameter file.")
            
        DatasetSeries.__init__(self, outputs=init_outputs, parallel=parallel,
                                setup_function=setup_function,
                                unit_base=self.unit_base)
        mylog.info("%d outputs loaded into time series.", len(init_outputs))

    def _parse_parameter_file(self):
        """
        Parses the parameter file and establishes the various
        dictionaries.
        """

        self.unit_base = {}

        # Let's read the file
        lines = open(self.parameter_filename).readlines()
        comments = ["%", ";"]
        for line in (l.strip() for l in lines):
            for comment in comments:
                if comment in line: line = line[0:line.find(comment)]
            if len(line) < 2: continue
            param, vals = (i.strip() for i in line.split(None, 1))
            # First we try to decipher what type of value it is.
            vals = vals.split()
            # Special case approaching.
            if "(do" in vals: vals = vals[:1]
            if len(vals) == 0:
                pcast = str # Assume NULL output
            else:
                v = vals[0]
                # Figure out if it's castable to floating point:
                try:
                    float(v)
                except ValueError:
                    pcast = str
                else:
                    if any("." in v or "e" in v for v in vals):
                        pcast = float
                    elif v == "inf":
                        pcast = str
                    else:
                        pcast = int
            # Now we figure out what to do with it.
            if param.startswith("Unit"):
                self.unit_base[param] = float(vals[0])
            if len(vals) == 0:
                vals = ""
            elif len(vals) == 1:
                vals = pcast(vals[0])
            else:
                vals = np.array([pcast(i) for i in vals])

            self.parameters[param] = vals

        # Domain dimensions for Gadget datasets are always 2x2x2 for octree
        self.domain_dimensions = np.array([2,2,2])

        if self.parameters["ComovingIntegrationOn"]:
            cosmo_attr = {"box_size": "BoxSize",
                          "omega_lambda": "OmegaLambda",
                          "omega_matter": "Omega0",
                          "hubble_constant": "HubbleParam"}
            self.initial_redshift = 1.0 / self.parameters["TimeBegin"] - 1.0
            self.final_redshift = 1.0 / self.parameters["TimeMax"] - 1.0
            self.cosmological_simulation = 1
            for a, v in cosmo_attr.items():
                if v not in self.parameters:
                    raise MissingParameter(self.parameter_filename, v)
                setattr(self, a, self.parameters[v])
            self.domain_left_edge = np.array([0., 0., 0.])
            self.domain_right_edge = np.array([1., 1., 1.]) * self.parameters['BoxSize']
        else:
            self.cosmological_simulation = 0
            self.omega_lambda = self.omega_matter = \
                self.hubble_constant = 0.0

    def _find_data_dir(self):
        """
        Find proper location for datasets.  First look where parameter file
        points, but if this doesn't exist then default to the current 
        directory.
        """
        if self.parameters["OutputDir"].startswith("/"):
            data_dir = self.parameters["OutputDir"]
        else:
            data_dir = os.path.join(self.directory,
                                    self.parameters["OutputDir"])
        if not os.path.exists(data_dir):
            mylog.info("OutputDir not found at %s, instead using %s." % 
                       (data_dir, self.directory))
            data_dir = self.directory
        self.data_dir = data_dir

    def _snapshot_format(self, index=None):
        """
        The snapshot filename for a given index.  Modify this for different 
        naming conventions.
        """

        if self.parameters["NumFilesPerSnapshot"] > 1:
            suffix = ".0"
        else:
            suffix = ""
        if self.parameters["SnapFormat"] == 3:
            suffix += ".hdf5"
        if index is None:
            count = "*"
        else:
            count = "%03d" % index
        filename = "%s_%s%s" % (self.parameters["SnapshotFileBase"],
                                count, suffix)
        return os.path.join(self.data_dir, filename)
                
    def _get_all_outputs(self, find_outputs=False):
        """
        Get all potential datasets and combine into a time-sorted list.
        """

        # Find the data directory where the outputs are
        self._find_data_dir()

        # Create the set of outputs from which further selection will be done.
        if find_outputs:
            self._find_outputs()
        else:
            if self.parameters["OutputListOn"]:
                a_values = [float(a) for a in 
                            open(os.path.join(self.data_dir, 
                                 self.parameters["OutputListFilename"]), 
                            "r").readlines()]
            else:
                a_values = [float(self.parameters["TimeOfFirstSnapshot"])]
                time_max = float(self.parameters["TimeMax"])
                while a_values[-1] < time_max:
                    if self.cosmological_simulation:
                        a_values.append(
                            a_values[-1] * self.parameters["TimeBetSnapshot"])
                    else:
                        a_values.append(
                            a_values[-1] + self.parameters["TimeBetSnapshot"])
                if a_values[-1] > time_max:
                    a_values[-1] = time_max

            if self.cosmological_simulation:
                self.all_outputs = \
                  [{"filename": self._snapshot_format(i),
                    "redshift": (1. / a - 1)}
                   for i, a in enumerate(a_values)]
                
                # Calculate times for redshift outputs.
                for output in self.all_outputs:
                    output["time"] = self.cosmology.t_from_z(output["redshift"])
            else:
                self.all_outputs = \
                  [{"filename": self._snapshot_format(i),
                    "time": self.quan(a, "code_time")}
                   for i, a in enumerate(a_values)]

            self.all_outputs.sort(key=lambda obj:obj["time"].to_ndarray())

    def _calculate_simulation_bounds(self):
        """
        Figure out the starting and stopping time and redshift for the simulation.
        """

        # Convert initial/final redshifts to times.
        if self.cosmological_simulation:
            self.initial_time = self.cosmology.t_from_z(self.initial_redshift)
            self.initial_time.units.registry = self.unit_registry
            self.final_time = self.cosmology.t_from_z(self.final_redshift)
            self.final_time.units.registry = self.unit_registry

        # If not a cosmology simulation, figure out the stopping criteria.
        else:
            if "TimeBegin" in self.parameters:
                self.initial_time = self.quan(self.parameters["TimeBegin"], "code_time")
            else:
                self.initial_time = self.quan(0., "code_time")

            if "TimeMax" in self.parameters:
                self.final_time = self.quan(self.parameters["TimeMax"], "code_time")
            else:
                self.final_time = None
            if "TimeMax" not in self.parameters:
                raise NoStoppingCondition(self.parameter_filename)

    def _find_outputs(self):
        """
        Search for directories matching the data dump keywords.
        If found, get dataset times py opening the ds.
        """
        potential_outputs = glob.glob(self._snapshot_format())
        self.all_outputs = self._check_for_outputs(potential_outputs)
        self.all_outputs.sort(key=lambda obj: obj["time"])
        only_on_root(mylog.info, "Located %d total outputs.", len(self.all_outputs))

        # manually set final time and redshift with last output
        if self.all_outputs:
            self.final_time = self.all_outputs[-1]["time"]
            if self.cosmological_simulation:
                self.final_redshift = self.all_outputs[-1]["redshift"]

    def _check_for_outputs(self, potential_outputs):
        r"""
        Check a list of files to see if they are valid datasets.
        """

        only_on_root(mylog.info, "Checking %d potential outputs.", 
                     len(potential_outputs))

        my_outputs = {}
        for my_storage, output in parallel_objects(potential_outputs, 
                                                   storage=my_outputs):
            if os.path.exists(output):
                try:
                    ds = load(output)
                    if ds is not None:
                        my_storage.result = {"filename": output,
                                             "time": ds.current_time.in_units("s")}
                        if ds.cosmological_simulation:
                            my_storage.result["redshift"] = ds.current_redshift
                except YTOutputNotIdentified:
                    mylog.error("Failed to load %s", output)
        my_outputs = [my_output for my_output in my_outputs.values() \
                      if my_output is not None]
        return my_outputs

    def _write_cosmology_outputs(self, filename, outputs, start_index,
                                 decimals=3):
        r"""
        Write cosmology output parameters for a cosmology splice.
        """

        mylog.info("Writing redshift output list to %s.", filename)
        f = open(filename, "w")
        for output in outputs:
            f.write("%f\n" % (1. / (1. + output["redshift"])))
        f.close()



"""
Tests for callbacks



"""
from __future__ import absolute_import

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------
import tempfile
import shutil
from numpy.testing import \
    assert_raises

from yt.config import \
    ytcfg
from yt.testing import \
    fake_amr_ds, \
    fake_tetrahedral_ds, \
    fake_hexahedral_ds, \
    assert_fname
import yt.units as u
from yt.utilities.exceptions import \
    YTPlotCallbackError, \
    YTDataTypeUnsupported
from yt.visualization.api import \
    SlicePlot, ProjectionPlot, OffAxisSlicePlot
import contextlib

# These are a very simple set of tests that verify that each callback is or is
# not working.  They only check that it functions without an error; they do not
# check that it is providing correct results.

# These are the callbacks still to test:
#
#  X velocity
#  X magnetic_field
#  X quiver
#  X contour
#  X grids
#    streamlines
#    units
#  X line
#    cquiver
#    clumps
#  X arrow
#  X marker
#  X sphere
#    hop_circles
#    hop_particles
#    coord_axes
#  X text
#  X particles
#    title
#    flash_ray_data
#  X timestamp
#  X scale
#    material_boundary
#  X ray
#  X line_integral_convolution

@contextlib.contextmanager
def _cleanup_fname():
    tmpdir = tempfile.mkdtemp()
    yield tmpdir
    shutil.rmtree(tmpdir)

def test_timestamp_callback():
    with _cleanup_fname() as prefix:
        ax = 'z'
        vector = [1.0,1.0,1.0]
        ds = fake_amr_ds(fields = ("density",))
        p = ProjectionPlot(ds, ax, "density")
        p.annotate_timestamp()
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, ax, "density")
        p.annotate_timestamp()
        assert_fname(p.save(prefix)[0])
        p = OffAxisSlicePlot(ds, vector, "density")
        p.annotate_timestamp()
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", "density")
        p.annotate_timestamp(corner='lower_right', redshift=True,
                             draw_inset_box=True)
        p.save(prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = ("density",), geometry="spherical")
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_timestamp(coord_system="data")
        assert_raises(YTDataTypeUnsupported, p.save, prefix)
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_timestamp(coord_system="axis")
        assert_fname(p.save(prefix)[0])

def test_scale_callback():
    with _cleanup_fname() as prefix:
        ax = 'z'
        vector = [1.0,1.0,1.0]
        ds = fake_amr_ds(fields = ("density",))
        p = ProjectionPlot(ds, ax, "density")
        p.annotate_scale()
        assert_fname(p.save(prefix)[0])
        p = ProjectionPlot(ds, ax, "density", width=(0.5, 1.0))
        p.annotate_scale()
        assert_fname(p.save(prefix)[0])
        p = ProjectionPlot(ds, ax, "density", width=(1.0, 1.5))
        p.annotate_scale()
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, ax, "density")
        p.annotate_scale()
        assert_fname(p.save(prefix)[0])
        p = OffAxisSlicePlot(ds, vector, "density")
        p.annotate_scale()
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", "density")
        p.annotate_scale(corner='upper_right', coeff=10., unit='kpc')
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, "x", "density")
        p.annotate_scale(text_args={"size": 24})
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, "x", "density")
        p.annotate_scale(text_args={"font": 24})
        assert_raises(YTPlotCallbackError)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = ("density",), geometry="spherical")
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_scale()
        assert_raises(YTDataTypeUnsupported, p.save, prefix)
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_scale(coord_system="axis")
        assert_raises(YTDataTypeUnsupported, p.save, prefix)

def test_line_callback():
    with _cleanup_fname() as prefix:
        ax = 'z'
        vector = [1.0,1.0,1.0]
        ds = fake_amr_ds(fields = ("density",))
        p = ProjectionPlot(ds, ax, "density")
        p.annotate_line([0.1,0.1,0.1],[0.5,0.5,0.5])
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, ax, "density")
        p.annotate_line([0.1,0.1,0.1],[0.5,0.5,0.5])
        assert_fname(p.save(prefix)[0])
        p = OffAxisSlicePlot(ds, vector, "density")
        p.annotate_line([0.1,0.1,0.1],[0.5,0.5,0.5])
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", "density")
        p.annotate_line([0.1,0.1],[0.5,0.5], coord_system='axis',
                        plot_args={'color':'red'})
        p.save(prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = ("density",), geometry="spherical")
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_line([0.1,0.1,0.1],[0.5,0.5,0.5])
        assert_raises(YTDataTypeUnsupported, p.save, prefix)
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_line([0.1,0.1],[0.5,0.5], coord_system="axis")
        assert_fname(p.save(prefix)[0])

def test_ray_callback():
    with _cleanup_fname() as prefix:
        ax = 'z'
        vector = [1.0,1.0,1.0]
        ds = fake_amr_ds(fields = ("density",))
        ray = ds.ray((0.1, 0.2, 0.3), (1.6, 1.8, 1.5))
        oray = ds.ortho_ray(0, (0.3, 0.4))
        p = ProjectionPlot(ds, ax, "density")
        p.annotate_ray(oray)
        p.annotate_ray(ray)
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, ax, "density")
        p.annotate_ray(oray)
        p.annotate_ray(ray)
        assert_fname(p.save(prefix)[0])
        p = OffAxisSlicePlot(ds, vector, "density")
        p.annotate_ray(oray)
        p.annotate_ray(ray)
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", "density")
        p.annotate_ray(oray)
        p.annotate_ray(ray, plot_args={'color':'red'})
        p.save(prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = ("density",), geometry="spherical")
        ray = ds.ray((0.1, 0.2, 0.3), (1.6, 1.8, 1.5))
        oray = ds.ortho_ray(0, (0.3, 0.4))
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_ray(oray)
        assert_raises(YTDataTypeUnsupported, p.save, prefix)
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_ray(ray)
        assert_raises(YTDataTypeUnsupported, p.save, prefix)

def test_arrow_callback():
    with _cleanup_fname() as prefix:
        ax = 'z'
        vector = [1.0,1.0,1.0]
        ds = fake_amr_ds(fields = ("density",))
        p = ProjectionPlot(ds, ax, "density")
        p.annotate_arrow([0.5,0.5,0.5])
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, ax, "density")
        p.annotate_arrow([0.5,0.5,0.5])
        assert_fname(p.save(prefix)[0])
        p = OffAxisSlicePlot(ds, vector, "density")
        p.annotate_arrow([0.5,0.5,0.5])
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", "density")
        p.annotate_arrow([0.5,0.5], coord_system='axis', length=0.05)
        p.save(prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = ("density",), geometry="spherical")
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_arrow([0.5,0.5,0.5])
        assert_raises(YTDataTypeUnsupported, p.save, prefix)
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_arrow([0.5,0.5], coord_system="axis")
        assert_fname(p.save(prefix)[0])

def test_marker_callback():
    with _cleanup_fname() as prefix:
        ax = 'z'
        vector = [1.0,1.0,1.0]
        ds = fake_amr_ds(fields = ("density",))
        p = ProjectionPlot(ds, ax, "density")
        p.annotate_marker([0.5,0.5,0.5])
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, ax, "density")
        p.annotate_marker([0.5,0.5,0.5])
        assert_fname(p.save(prefix)[0])
        p = OffAxisSlicePlot(ds, vector, "density")
        p.annotate_marker([0.5,0.5,0.5])
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", "density")
        p.annotate_marker([0.5,0.5], coord_system='axis', marker='*')
        p.save(prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = ("density",), geometry="spherical")
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_marker([0.5,0.5,0.5])
        assert_raises(YTDataTypeUnsupported, p.save, prefix)
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_marker([0.5,0.5], coord_system="axis")
        assert_fname(p.save(prefix)[0])

def test_particles_callback():
    with _cleanup_fname() as prefix:
        ax = 'z'
        ds = fake_amr_ds(fields=("density",), particles=1)
        p = ProjectionPlot(ds, ax, "density")
        p.annotate_particles((10, "Mpc"))
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, ax, "density")
        p.annotate_particles((10, "Mpc"))
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", "density")
        p.annotate_particles((10, "Mpc"), p_size=1.0, col="k", marker="o",
                             stride=1, ptype="all", minimum_mass=None,
                             alpha=1.0)
        p.save(prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields=("density",), geometry="spherical")
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_particles((10, "Mpc"))
        assert_raises(YTDataTypeUnsupported, p.save, prefix)

def test_sphere_callback():
    with _cleanup_fname() as prefix:
        ax = 'z'
        vector = [1.0,1.0,1.0]
        ds = fake_amr_ds(fields = ("density",))
        p = ProjectionPlot(ds, ax, "density")
        p.annotate_sphere([0.5,0.5,0.5], 0.1)
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, ax, "density")
        p.annotate_sphere([0.5,0.5,0.5], 0.1)
        assert_fname(p.save(prefix)[0])
        p = OffAxisSlicePlot(ds, vector, "density")
        p.annotate_sphere([0.5,0.5,0.5], 0.1)
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", "density")
        p.annotate_sphere([0.5,0.5], 0.1, coord_system='axis', text='blah')
        p.save(prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = ("density",), geometry="spherical")
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_sphere([0.5,0.5,0.5], 0.1)
        assert_raises(YTDataTypeUnsupported, p.save, prefix)
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_sphere([0.5,0.5], 0.1, coord_system='axis', text='blah')
        assert_fname(p.save(prefix)[0])

def test_text_callback():
    with _cleanup_fname() as prefix:
        ax = 'z'
        vector = [1.0,1.0,1.0]
        ds = fake_amr_ds(fields = ("density",))
        p = ProjectionPlot(ds, ax, "density")
        p.annotate_text([0.5,0.5,0.5], 'dinosaurs!')
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, ax, "density")
        p.annotate_text([0.5,0.5,0.5], 'dinosaurs!')
        assert_fname(p.save(prefix)[0])
        p = OffAxisSlicePlot(ds, vector, "density")
        p.annotate_text([0.5,0.5,0.5], 'dinosaurs!')
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", "density")
        p.annotate_text([0.5,0.5], 'dinosaurs!', coord_system='axis',
                        text_args={'color':'red'})
        p.save(prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = ("density",), geometry="spherical")
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_text([0.5,0.5,0.5], 'dinosaurs!')
        assert_raises(YTDataTypeUnsupported, p.save, prefix)
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_text([0.5,0.5], 'dinosaurs!', coord_system='axis',
                        text_args={'color':'red'})
        assert_fname(p.save(prefix)[0])

def test_velocity_callback():
    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields =
            ("density", "velocity_x", "velocity_y", "velocity_z"))
        for ax in 'xyz':
            p = ProjectionPlot(ds, ax, "density", weight_field="density")
            p.annotate_velocity()
            assert_fname(p.save(prefix)[0])
            p = SlicePlot(ds, ax, "density")
            p.annotate_velocity()
            assert_fname(p.save(prefix)[0])
        # Test for OffAxis Slice
        p = SlicePlot(ds, [1, 1, 0], 'density', north_vector=[0, 0, 1])
        p.annotate_velocity(factor=40, normalize=True)
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", "density")
        p.annotate_velocity(factor=8, scale=0.5, scale_units="inches",
                            normalize = True)
        assert_fname(p.save(prefix)[0])

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = 
            ("density", "velocity_r", "velocity_theta", "velocity_phi"),
            geometry="spherical")
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_velocity(factor=40, normalize=True)
        assert_raises(YTDataTypeUnsupported, p.save, prefix)

def test_magnetic_callback():
    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = ("density", "magnetic_field_x",
          "magnetic_field_y", "magnetic_field_z"))
        for ax in 'xyz':
            p = ProjectionPlot(ds, ax, "density", weight_field="density")
            p.annotate_magnetic_field()
            assert_fname(p.save(prefix)[0])
            p = SlicePlot(ds, ax, "density")
            p.annotate_magnetic_field()
            assert_fname(p.save(prefix)[0])
        # Test for OffAxis Slice
        p = SlicePlot(ds, [1, 1, 0], 'density', north_vector=[0, 0, 1])
        p.annotate_magnetic_field(factor=40, normalize=True)
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", "density")
        p.annotate_magnetic_field(factor=8, scale=0.5,
            scale_units="inches", normalize = True)
        assert_fname(p.save(prefix)[0])

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = ("density", "magnetic_field_r",
          "magnetic_field_theta", "magnetic_field_phi"),
          geometry="spherical")
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_magnetic_field(factor=8, scale=0.5,
            scale_units="inches", normalize = True)
        assert_raises(YTDataTypeUnsupported, p.save, prefix)

def test_quiver_callback():
    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields =
            ("density", "velocity_x", "velocity_y", "velocity_z"))
        for ax in 'xyz':
            p = ProjectionPlot(ds, ax, "density")
            p.annotate_quiver("velocity_x", "velocity_y")
            assert_fname(p.save(prefix)[0])
            p = ProjectionPlot(ds, ax, "density", weight_field="density")
            p.annotate_quiver("velocity_x", "velocity_y")
            assert_fname(p.save(prefix)[0])
            p = SlicePlot(ds, ax, "density")
            p.annotate_quiver("velocity_x", "velocity_y")
            assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", "density")
        p.annotate_quiver("velocity_x", "velocity_y", factor=8, scale=0.5,
            scale_units="inches", normalize = True,
            bv_x = 0.5 * u.cm / u.s,
            bv_y = 0.5 * u.cm / u.s)
        assert_fname(p.save(prefix)[0])

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = 
            ("density", "velocity_x", "velocity_theta", "velocity_phi"),
            geometry="spherical")
        p = ProjectionPlot(ds, "r", "density")
        p.annotate_quiver("velocity_theta", "velocity_phi", factor=8, scale=0.5,
            scale_units="inches", normalize = True,
            bv_x = 0.5 * u.cm / u.s,
            bv_y = 0.5 * u.cm / u.s)
        assert_raises(YTDataTypeUnsupported, p.save, prefix)

def test_contour_callback():
    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = ("density", "temperature"))
        for ax in 'xyz':
            p = ProjectionPlot(ds, ax, "density")
            p.annotate_contour("temperature")
            assert_fname(p.save(prefix)[0])
            p = ProjectionPlot(ds, ax, "density", weight_field="density")
            p.annotate_contour("temperature")
            assert_fname(p.save(prefix)[0])
            p = SlicePlot(ds, ax, "density")
            p.annotate_contour("temperature") # BREAKS WITH ndarray
            assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", "density")
        p.annotate_contour("temperature", ncont=10, factor=8,
            take_log=False, clim=(0.4, 0.6),
            plot_args={'lw':2.0}, label=True,
            text_args={'text-size':'x-large'})
        p.save(prefix)

        p = SlicePlot(ds, "x", "density")
        s2 = ds.slice(0, 0.2)
        p.annotate_contour("temperature", ncont=10, factor=8,
            take_log=False, clim=(0.4, 0.6),
            plot_args={'lw':2.0}, label=True,
            text_args={'text-size':'x-large'},
            data_source=s2)
        p.save(prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = ("density", "temperature"),
                         geometry="spherical")
        p = SlicePlot(ds, "r", "density")
        p.annotate_contour("temperature", ncont=10, factor=8,
            take_log=False, clim=(0.4, 0.6),
            plot_args={'lw':2.0}, label=True,
            text_args={'text-size':'x-large'})
        assert_raises(YTDataTypeUnsupported, p.save, prefix)


def test_grids_callback():
    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = ("density",))
        for ax in 'xyz':
            p = ProjectionPlot(ds, ax, "density")
            p.annotate_grids()
            assert_fname(p.save(prefix)[0])
            p = ProjectionPlot(ds, ax, "density", weight_field="density")
            p.annotate_grids()
            assert_fname(p.save(prefix)[0])
            p = SlicePlot(ds, ax, "density")
            p.annotate_grids()
            assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", "density")
        p.annotate_grids(alpha=0.7, min_pix=10, min_pix_ids=30,
            draw_ids=True, periodic=False, min_level=2,
            max_level=3, cmap="gist_stern")
        p.save(prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = ("density",), geometry="spherical")
        p = SlicePlot(ds, "r", "density")
        p.annotate_grids(alpha=0.7, min_pix=10, min_pix_ids=30,
            draw_ids=True, periodic=False, min_level=2,
            max_level=3, cmap="gist_stern")
        assert_raises(YTDataTypeUnsupported, p.save, prefix)


def test_cell_edges_callback():
    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = ("density",))
        for ax in 'xyz':
            p = ProjectionPlot(ds, ax, "density")
            p.annotate_cell_edges()
            assert_fname(p.save(prefix)[0])
            p = ProjectionPlot(ds, ax, "density", weight_field="density")
            p.annotate_cell_edges()
            assert_fname(p.save(prefix)[0])
            p = SlicePlot(ds, ax, "density")
            p.annotate_cell_edges()
            assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", "density")
        p.annotate_cell_edges(alpha=0.7, line_width=0.9,
                              color=(0.0, 1.0, 1.0))
        p.save(prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = ("density",), geometry="spherical")
        p = SlicePlot(ds, "r", "density")
        p.annotate_cell_edges()
        assert_raises(YTDataTypeUnsupported, p.save, prefix)

def test_mesh_lines_callback():
    with _cleanup_fname() as prefix:

        ds = fake_hexahedral_ds()
        for field in ds.field_list:
            sl = SlicePlot(ds, 1, field)
            sl.annotate_mesh_lines(plot_args={'color':'black'})
            assert_fname(sl.save(prefix)[0])

        ds = fake_tetrahedral_ds()
        for field in ds.field_list:
            sl = SlicePlot(ds, 1, field)
            sl.annotate_mesh_lines(plot_args={'color':'black'})
            assert_fname(sl.save(prefix)[0])
                

def test_line_integral_convolution_callback():
    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields =
            ("density", "velocity_x", "velocity_y", "velocity_z"))
        for ax in 'xyz':
            p = ProjectionPlot(ds, ax, "density")
            p.annotate_line_integral_convolution("velocity_x", "velocity_y")
            assert_fname(p.save(prefix)[0])
            p = ProjectionPlot(ds, ax, "density", weight_field="density")
            p.annotate_line_integral_convolution("velocity_x", "velocity_y")
            assert_fname(p.save(prefix)[0])
            p = SlicePlot(ds, ax, "density")
            p.annotate_line_integral_convolution("velocity_x", "velocity_y")
            assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", "density")
        p.annotate_line_integral_convolution("velocity_x", "velocity_y",
                                             kernellen=100., lim=(0.4,0.7),
                                             cmap=ytcfg.get("yt", "default_colormap"),
                                             alpha=0.9, const_alpha=True)
        p.save(prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields = 
            ("density", "velocity_r", "velocity_theta", "velocity_phi"),
            geometry="spherical")
        p = SlicePlot(ds, "r", "density")
        p.annotate_line_integral_convolution("velocity_theta", "velocity_phi")
        assert_raises(YTDataTypeUnsupported, p.save, prefix)



"""
ARTIO frontend tests




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.testing import \
    assert_equal, \
    requires_file, \
    units_override_check
from yt.utilities.answer_testing.framework import \
    requires_ds, \
    data_dir_load, \
    PixelizedProjectionValuesTest, \
    FieldValuesTest, \
    create_obj
from yt.frontends.artio.api import ARTIODataset

_fields = ("temperature", "density", "velocity_magnitude",
           ("deposit", "all_density"), ("deposit", "all_count"))

sizmbhloz = "sizmbhloz-clref04SNth-rs9_a0.9011/sizmbhloz-clref04SNth-rs9_a0.9011.art"
@requires_ds(sizmbhloz)
def test_sizmbhloz():
    ds = data_dir_load(sizmbhloz)
    ds.max_range = 1024*1024
    assert_equal(str(ds), "sizmbhloz-clref04SNth-rs9_a0.9011.art")
    dso = [ None, ("sphere", ("max", (0.1, 'unitary')))]
    for dobj_name in dso:
        for field in _fields:
            for axis in [0, 1, 2]:
                for weight_field in [None, "density"]:
                    yield PixelizedProjectionValuesTest(
                        ds, axis, field, weight_field,
                        dobj_name)
            yield FieldValuesTest(ds, field, dobj_name)
        dobj = create_obj(ds, dobj_name)
        s1 = dobj["ones"].sum()
        s2 = sum(mask.sum() for block, mask in dobj.blocks)
        assert_equal(s1, s2)
    assert_equal(ds.particle_type_counts, {'N-BODY': 100000, 'STAR': 110650})


@requires_file(sizmbhloz)
def test_ARTIODataset():
    assert isinstance(data_dir_load(sizmbhloz), ARTIODataset)

@requires_file(sizmbhloz)
def test_units_override():
    units_override_check(sizmbhloz)



"""
Eagle frontend tests using the snapshot_028_z000p000 dataset




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2015, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.testing import \
    requires_file
from yt.utilities.answer_testing.framework import \
    data_dir_load
from yt.frontends.eagle.api import EagleDataset

s28 = "snapshot_028_z000p000/snap_028_z000p000.0.hdf5"
@requires_file(s28)
def test_EagleDataset():
    assert isinstance(data_dir_load(s28), EagleDataset)

s399 = "snipshot_399_z000p000/snip_399_z000p000.0.hdf5"
@requires_file(s399)
def test_Snipshot():
    ds = data_dir_load(s399)
    ds.index
    assert isinstance(ds, EagleDataset)



#-----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from yt.testing import \
    assert_allclose_units, \
    fake_random_ds

def test_vector_component_conversions():
    ds = fake_random_ds(16)

    ad = ds.all_data()

    vmag = ad['velocity_magnitude']
    vmag_cart = np.sqrt(
        ad['velocity_x']**2 +
        ad['velocity_y']**2 +
        ad['velocity_z']**2
    )

    assert_allclose_units(vmag, vmag_cart)

    vmag_cyl = np.sqrt(
        ad['velocity_cylindrical_radius']**2 +
        ad['velocity_cylindrical_theta']**2 +
        ad['velocity_cylindrical_z']**2
    )

    assert_allclose_units(vmag, vmag_cyl)

    vmag_sph = np.sqrt(
        ad['velocity_spherical_radius']**2 +
        ad['velocity_spherical_theta']**2 +
        ad['velocity_spherical_phi']**2
    )

    assert_allclose_units(vmag, vmag_sph)



"""
API for Gadget frontend




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2014-2015, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from .data_structures import \
    GadgetDataset, \
    GadgetHDF5Dataset, \
    GadgetFieldInfo

from .io import \
    IOHandlerGadgetBinary, \
    IOHandlerGadgetHDF5

from .simulation_handling import \
    GadgetSimulation

from . import tests



from __future__ import print_function
from yt.geometry.fake_octree import create_fake_octree
from yt.geometry.oct_container import RAMSESOctreeContainer, ParticleOctreeContainer
import numpy as np

nocts= 3
max_level = 12
dn = 2
dd = np.ones(3,dtype='i4')*dn
dle = np.ones(3,dtype='f8')*0.0
dre = np.ones(3,dtype='f8')
fsub = 0.25
domain = 1

oct_handler = RAMSESOctreeContainer(dd,dle,dre)
leaves = create_fake_octree(oct_handler, nocts, max_level, dd, dle, dre, fsub)
mask = np.ones((nocts,8),dtype='bool')
cell_count = nocts*8
oct_counts = oct_handler.count_levels(max_level, 1, mask)
level_counts = np.concatenate(([0,],np.cumsum(oct_counts)))
fc = oct_handler.fcoords(domain,mask,cell_count, level_counts.copy())
leavesb = oct_handler.count_leaves(mask)
assert leaves == leavesb

#Now take the fcoords, call them particles and recreate the same octree
print("particle-based recreate")
oct_handler2 = ParticleOctreeContainer(dd,dle,dre)
oct_handler2.allocate_domains([nocts])
oct_handler2.n_ref = 1 #specifically make a maximum of 1 particle per oct
oct_handler2.add(fc, 1)
print("added particles")
cell_count2 = nocts*8
oct_counts2 = oct_handler.count_levels(max_level, 1, mask)
level_counts2 = np.concatenate(([0,],np.cumsum(oct_counts)))
fc2 = oct_handler.fcoords(domain,mask,cell_count, level_counts.copy())
leaves2 = oct_handler2.count_leaves(mask)
assert leaves == leaves2

print("success")



"""
Halo object.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

class Halo(object):
    particles = None
    def __init__(self, halo_catalog):
        self.halo_catalog = halo_catalog
        self.quantities = {}



cdef void Q1Function3D(double* fx,
                       double* x,
                       double* vertices,
                       double* phys_x) nogil 

 
cdef void Q1Jacobian3D(double* rcol,
                       double* scol,
                       double* tcol,
                       double* x,
                       double* vertices,
                       double* phys_x) nogil 

 
cdef void Q1Function2D(double* fx,
                       double* x,
                       double* vertices,
                       double* phys_x) nogil 

 
cdef void Q1Jacobian2D(double* rcol,
                       double* scol,
                       double* x,
                       double* vertices,
                       double* phys_x) nogil 

 
cdef void Q2Function2D(double* fx,
                       double* x,
                       double* vertices,
                       double* phys_x) nogil 

 
cdef void Q2Jacobian2D(double* rcol,
                       double* scol,
                       double* x,
                       double* vertices,
                       double* phys_x) nogil 

 
cdef void Tet2Function3D(double* fx,
                       double* x,
                       double* vertices,
                       double* phys_x) nogil 

 
cdef void Tet2Jacobian3D(double* rcol,
                       double* scol,
                       double* tcol,
                       double* x,
                       double* vertices,
                       double* phys_x) nogil 

 
cdef void T2Function2D(double* fx,
                       double* x,
                       double* vertices,
                       double* phys_x) nogil 

 
cdef void T2Jacobian2D(double* rcol,
                       double* scol,
                       double* x,
                       double* vertices,
                       double* phys_x) nogil 

 
cdef void W1Function3D(double* fx,
                       double* x,
                       double* vertices,
                       double* phys_x) nogil 

 
cdef void W1Jacobian3D(double* rcol,
                       double* scol,
                       double* tcol,
                       double* x,
                       double* vertices,
                       double* phys_x) nogil 

 



"""
ART-specific fields



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.fields.field_info_container import \
    FieldInfoContainer

b_units = "code_magnetic"
ra_units = "code_length / code_time**2"
rho_units = "code_mass / code_length**3"
vel_units = "code_velocity"
# NOTE: ARTIO uses momentum density.
mom_units = "code_mass / (code_length**2 * code_time)"
en_units = "code_mass*code_velocity**2/code_length**3"

class ARTFieldInfo(FieldInfoContainer):
    known_other_fields = (
        ("Density", (rho_units, ["density"], None)),
        ("TotalEnergy", (en_units, ["total_energy"], None)),
        ("XMomentumDensity", (mom_units, ["momentum_x"], None)),
        ("YMomentumDensity", (mom_units, ["momentum_y"], None)),
        ("ZMomentumDensity", (mom_units, ["momentum_z"], None)),
        ("Pressure", ("", ["pressure"], None)), # Unused
        ("Gamma", ("", ["gamma"], None)),
        ("GasEnergy", (en_units, ["thermal_energy"], None)),
        ("MetalDensitySNII", (rho_units, ["metal_ii_density"], None)),
        ("MetalDensitySNIa", (rho_units, ["metal_ia_density"], None)),
        ("PotentialNew", ("", ["potential"], None)),
        ("PotentialOld", ("", ["gas_potential"], None)),
    )

    known_particle_fields = (
        ("particle_position_x", ("code_length", [], None)),
        ("particle_position_y", ("code_length", [], None)),
        ("particle_position_z", ("code_length", [], None)),
        ("particle_velocity_x", (vel_units, [], None)),
        ("particle_velocity_y", (vel_units, [], None)),
        ("particle_velocity_z", (vel_units, [], None)),
        ("particle_mass", ("code_mass", [], None)),
        ("particle_index", ("", [], None)),
        ("particle_species", ("", ["particle_type"], None)),
        ("particle_creation_time", ("Gyr", [], None)),
        ("particle_mass_initial", ("code_mass", [], None)),
        ("particle_metallicity1", ("", [], None)),
        ("particle_metallicity2", ("", [], None)),
    )

    def setup_fluid_fields(self):
        unit_system = self.ds.unit_system
        def _temperature(field, data):
            r0 = data.ds.parameters['boxh'] / data.ds.parameters['ng']
            tr = data.ds.quan(3.03e5 * r0**2, 'K/code_velocity**2')
            tr *= data.ds.parameters['wmu'] * data.ds.parameters['Om0']
            tr *= (data.ds.parameters['gamma'] - 1.)
            tr /= data.ds.parameters['aexpn']**2
            return tr * data['art', 'GasEnergy'] / data['art', 'Density']
        self.add_field(('gas', 'temperature'), sampling_type="cell", 
                       function=_temperature, 
                       units=unit_system["temperature"])

        def _get_vel(axis):
            def velocity(field, data):
                return (data[('gas','momentum_%s' % axis)] /
                        data[('gas','density')])
            return velocity
        for ax in 'xyz':
            self.add_field(('gas','velocity_%s' % ax), sampling_type="cell", 
                           function = _get_vel(ax),
                           units=unit_system["velocity"])

        def _momentum_magnitude(field, data):
            tr = (data['gas','momentum_x']**2 +
                  data['gas','momentum_y']**2 +
                  data['gas','momentum_z']**2)**0.5
            tr *= data['index','cell_volume'].in_units('cm**3')
            return tr
        self.add_field(('gas', 'momentum_magnitude'), sampling_type="cell", 
                       function=_momentum_magnitude,
                       units=unit_system["momentum"])

        def _velocity_magnitude(field, data):
            tr = data['gas','momentum_magnitude']
            tr /= data['gas','cell_mass']
            return tr
        self.add_field(('gas', 'velocity_magnitude'), sampling_type="cell", 
                       function=_velocity_magnitude,
                       units=unit_system["velocity"])

        def _metal_density(field, data):
            tr = data['gas','metal_ia_density']
            tr += data['gas','metal_ii_density']
            return tr
        self.add_field(('gas','metal_density'), sampling_type="cell", 
                       function=_metal_density,
                       units=unit_system["density"])

        def _metal_mass_fraction(field, data):
            tr = data['gas','metal_density']
            tr /= data['gas','density']
            return tr
        self.add_field(('gas', 'metal_mass_fraction'), sampling_type="cell", 
                       function=_metal_mass_fraction,
                       units='')

        def _H_mass_fraction(field, data):
            tr = (1. - data.ds.parameters['Y_p'] - 
                  data['gas', 'metal_mass_fraction'])
            return tr
        self.add_field(('gas', 'H_mass_fraction'), sampling_type="cell", 
                       function=_H_mass_fraction,
                       units='')

        def _metallicity(field, data):
            tr = data['gas','metal_mass_fraction']
            tr /= data['gas','H_mass_fraction']
            return tr
        self.add_field(('gas', 'metallicity'), sampling_type="cell", 
                       function=_metallicity,
                       units='')



import numpy as np
import os, sys
cimport numpy as np
cimport cython
from cython cimport floating
#from cpython.mem cimport PyMem_Malloc
from libc.stdlib cimport malloc, free
import sys

# Importing relevant rockstar data types particle, fof halo, halo

cdef import from "particle.h":
    struct particle:
        np.int64_t id
        float pos[6]

cdef import from "rockstar.h":
    particle *global_particles "p"
    void rockstar_cleanup()

cdef import from "fof.h":
    struct fof:
        np.int64_t num_p
        particle *particles

cdef import from "halo.h":
    struct halo:
        np.int64_t id
        float pos[6]
        float corevel[3]
        float bulkvel[3]
        float m, r, child_r, vmax_r, mgrav, vmax, rvmax, rs, klypin_rs, vrms
        float J[3]
        float energy, spin
        float alt_m[4]
        float Xoff, Voff, b_to_a, c_to_a
        float A[3]
        float b_to_a2, c_to_a2
        float A2[3]
        float bullock_spin, kin_to_pot, m_pe_b, m_pe_d
        np.int64_t num_p, num_child_particles, p_start, desc, flags, n_core
        float min_pos_err, min_vel_err, min_bulkvel_err, _pad

ctypedef struct haloflat:
    np.int64_t id
    float pos_x, pos_y, pos_z, vel_x, vel_y, vel_z
    float corevel_x, corevel_y, corevel_z
    float bulkvel_x, bulkvel_y, bulkvel_z
    float m, r, child_r, vmax_r, mgrav, vmax, rvmax, rs, klypin_rs, vrms
    float Jx, Jy, Jz
    float energy, spin
    float alt_m1, alt_m2, alt_m3, alt_m4
    float Xoff, Voff, b_to_a, c_to_a
    float Ax, Ay, Az
    float b_to_a2, c_to_a2, A2x, A2y, A2z
    float bullock_spin, kin_to_pot, m_pe_b, m_pe_d
    np.int64_t num_p, num_child_particles, p_start, desc, flags, n_core
    float min_pos_err, min_vel_err, min_bulkvel_err, _pad

# For finding sub halos import finder function and global variable
# rockstar uses to store the results

cdef import from "groupies.h":
    void find_subs(fof *f) nogil
    halo *halos
    np.int64_t num_halos
    void calc_mass_definition() nogil
    void free_particle_copies() nogil
    void alloc_particle_copies(np.int64_t total_copies) nogil
    void free_halos() nogil
    float max_halo_radius(halo *h) nogil

# global in groupies.c
cdef extern double particle_thresh_dens[5]

# For outputing halos, rockstar style

cdef import from "meta_io.h":
    void output_halos(np.int64_t id_offset, np.int64_t snap, np.int64_t chunk, float *bounds) nogil

# For setting up the configuration of rockstar

cdef import from "config.h":
    void setup_config() nogil
    void output_config(char *fn) nogil

cdef import from "distance.h":
    void init_cosmology() nogil

cdef import from "config_vars.h":
    # Rockstar cleverly puts all of the config variables inside a templated
    # definition of their vaiables.
    char *FILE_FORMAT
    np.float64_t PARTICLE_MASS

    char *MASS_DEFINITION
    char *MASS_DEFINITION2
    char *MASS_DEFINITION3
    char *MASS_DEFINITION4
    char *MASS_DEFINITION5
    np.int64_t STRICT_SO_MASSES
    np.int64_t MIN_HALO_OUTPUT_SIZE
    np.float64_t FORCE_RES
    np.float64_t FORCE_RES_PHYS_MAX

    np.float64_t SCALE_NOW
    np.float64_t h0
    np.float64_t Ol
    np.float64_t Om
    np.float64_t W0
    np.float64_t WA

    np.int64_t GADGET_ID_BYTES
    np.float64_t GADGET_MASS_CONVERSION
    np.float64_t GADGET_LENGTH_CONVERSION
    np.int64_t GADGET_SKIP_NON_HALO_PARTICLES
    np.int64_t RESCALE_PARTICLE_MASS

    np.int64_t PARALLEL_IO
    char *PARALLEL_IO_SERVER_ADDRESS
    char *PARALLEL_IO_SERVER_PORT
    np.int64_t PARALLEL_IO_WRITER_PORT
    char *PARALLEL_IO_SERVER_INTERFACE
    char *RUN_ON_SUCCESS

    char *INBASE
    char *FILENAME
    np.int64_t STARTING_SNAP
    np.int64_t RESTART_SNAP
    np.int64_t NUM_SNAPS
    np.int64_t NUM_BLOCKS
    np.int64_t NUM_READERS
    np.int64_t PRELOAD_PARTICLES
    char *SNAPSHOT_NAMES
    char *LIGHTCONE_ALT_SNAPS
    char *BLOCK_NAMES

    char *OUTBASE
    np.float64_t OVERLAP_LENGTH
    np.int64_t NUM_WRITERS
    np.int64_t FORK_READERS_FROM_WRITERS
    np.int64_t FORK_PROCESSORS_PER_MACHINE

    char *OUTPUT_FORMAT
    np.int64_t DELETE_BINARY_OUTPUT_AFTER_FINISHED
    np.int64_t FULL_PARTICLE_CHUNKS
    char *BGC2_SNAPNAMES

    np.int64_t SHAPE_ITERATIONS
    np.int64_t WEIGHTED_SHAPES
    np.int64_t BOUND_PROPS
    np.int64_t BOUND_OUT_TO_HALO_EDGE
    np.int64_t DO_MERGER_TREE_ONLY
    np.int64_t IGNORE_PARTICLE_IDS
    np.float64_t EXACT_LL_CALC
    np.float64_t TRIM_OVERLAP
    np.float64_t ROUND_AFTER_TRIM
    np.int64_t LIGHTCONE
    np.int64_t PERIODIC

    np.float64_t LIGHTCONE_ORIGIN[3]
    np.float64_t LIGHTCONE_ALT_ORIGIN[3]

    np.float64_t LIMIT_CENTER[3]
    np.float64_t LIMIT_RADIUS

    np.int64_t SWAP_ENDIANNESS
    np.int64_t GADGET_VARIANT
    np.int64_t ART_VARIANT

    np.float64_t FOF_FRACTION
    np.float64_t FOF_LINKING_LENGTH
    np.float64_t INITIAL_METRIC_SCALING
    np.float64_t INCLUDE_HOST_POTENTIAL_RATIO
    np.int64_t TEMPORAL_HALO_FINDING
    np.int64_t MIN_HALO_PARTICLES
    np.float64_t UNBOUND_THRESHOLD
    np.int64_t ALT_NFW_METRIC
    np.int64_t EXTRA_PROFILING

    np.int64_t TOTAL_PARTICLES
    np.float64_t BOX_SIZE
    np.int64_t OUTPUT_LEVELS
    np.float64_t DUMP_PARTICLES[3]

    np.float64_t AVG_PARTICLE_SPACING
    np.int64_t SINGLE_SNAP



cdef class RockstarGroupiesInterface:
    
    cdef public object ds
    cdef public object fof

    # For future use/consistency
    def __cinit__(self,ds):
        self.ds = ds

    def setup_rockstar(self,
                       particle_mass,
                       int periodic = 1, force_res = None,
                       int min_halo_size = 25, outbase = "None",
                       write_config = False,  exact_ll_calc = False,
                       lightcone = False, lightcone_origin = [0,0,0],
                       callbacks = None, unbound_threshold=None):
        global FILENAME, FILE_FORMAT, NUM_SNAPS, STARTING_SNAP, h0, Ol, Om
        global BOX_SIZE, PERIODIC, PARTICLE_MASS, NUM_BLOCKS, NUM_READERS
        global FORK_READERS_FROM_WRITERS, PARALLEL_IO_WRITER_PORT, NUM_WRITERS
        global rh, SCALE_NOW, OUTBASE, MIN_HALO_OUTPUT_SIZE
        global OVERLAP_LENGTH, TOTAL_PARTICLES, FORCE_RES
        global OUTPUT_FORMAT, EXTRA_PROFILING
        global STRICT_SO_MASSES, EXACT_LL_CALC
        global LIGHTCONE, LIGHTCONE_ORIGIN

        if force_res is not None:
            FORCE_RES=np.float64(force_res)

        OVERLAP_LENGTH = 0.0
        
        # Set to 0.0 if you plan on calculating spherical overdensity masses.
        # Otherwise filtering of halos in rockstar meta_io.c _should_print
        # will filter the wrong halos when halo mass is re-calculated before
        # output_halos
        global UNBOUND_THRESHOLD
        if unbound_threshold is not None:
            UNBOUND_THRESHOLD = unbound_threshold
        
        FILENAME = "inline.<block>"
        FILE_FORMAT = "GENERIC"
        OUTPUT_FORMAT = "BOTH"
        MIN_HALO_OUTPUT_SIZE=min_halo_size
        
        ds = self.ds

        h0 = ds.hubble_constant
        Ol = ds.omega_lambda
        Om = ds.omega_matter
        
        SCALE_NOW = 1.0/(ds.current_redshift+1.0)
        
        if not outbase =='None'.decode('UTF-8'):
            #output directory. since we can't change the output filenames
            #workaround is to make a new directory
            OUTBASE = outbase 

        PARTICLE_MASS = particle_mass.in_units('Msun/h')
        PERIODIC = periodic
        BOX_SIZE = ds.domain_width.in_units('Mpccm/h')[0]

        if exact_ll_calc: EXACT_LL_CALC = 1
        STRICT_SO_MASSES = 1    # presumably unused in our code path
        EXTRA_PROFILING = 0

        if lightcone:
            LIGHTCONE = 1
            LIGHTCONE_ORIGIN[0] = lightcone_origin[0]
            LIGHTCONE_ORIGIN[1] = lightcone_origin[1]
            LIGHTCONE_ORIGIN[2] = lightcone_origin[2]

        # Set up the configuration options
        setup_config()

        # Needs to be called so rockstar can use the particle mass parameter
        # to calculate virial quantities properly
        init_cosmology()
        calc_mass_definition()

        if write_config: output_config(NULL)

    def particle_thresh_dens(self):
        cdef np.ndarray d = np.array([particle_thresh_dens[0],
                                      particle_thresh_dens[1],
                                      particle_thresh_dens[2],
                                      particle_thresh_dens[3],
                                      particle_thresh_dens[4]],
                                     dtype=np.float64)
        return d

    def assign_masses(self, h, np.ndarray[np.float32_t, ndim=1] r, float force_res, \
                      double pmass, np.ndarray[np.float64_t, ndim=1] dens_thresh,
                      early_termination=False):
        """
        Assign spherical overdensity masses to halos.  r must be sorted

        Parameters
        ----------
        h: struct haloflat
            Assign masses to this halo
        r: np.ndarray
            Sorted array of particle radii
        force_res: float
            Force resolution, below which density is smoothed.
        dens_thresh: np.ndarray
            Thresholds for spherical overdensity mass calculation
        early_termination: bool
            Specifies whether or not to terminate mass calculation when
            first particle density is below the lowest density threshold.
            If False, may lead to overestimate of SO masses for subhalos,
            but gives a better comparison to plain rockstar masses with
            STRICT_SO=1. Default: False
        Returns
        -------
        None
        """
        cdef double total_mass = 0.0
        cdef double m = 0.0
        cdef double alt_m1 = 0.0
        cdef double alt_m2 = 0.0
        cdef double alt_m3 = 0.0
        cdef double alt_m4 = 0.0
        cdef double rr
        cdef double cur_dens
        cdef int min_ind = np.argmin(dens_thresh)
        cdef int eterm = early_termination
        for rr in r:
            if rr < force_res: rr = force_res
            total_mass += pmass
            cur_dens = total_mass/(rr*rr*rr)
            if cur_dens > dens_thresh[0]: m = total_mass
            if cur_dens > dens_thresh[1]: alt_m1 = total_mass
            if cur_dens > dens_thresh[2]: alt_m2 = total_mass
            if cur_dens > dens_thresh[3]: alt_m3 = total_mass
            if cur_dens > dens_thresh[4]: alt_m4 = total_mass
            if eterm and cur_dens <= dens_thresh[min_ind]:
                break
        h['m'] = m
        h['alt_m1'] = alt_m1
        h['alt_m2'] = alt_m2
        h['alt_m3'] = alt_m3
        h['alt_m4'] = alt_m4
        # if cur_dens > dens_thresh[1]:
            # This is usually a subhalo problem, and we don't know who is a subhalo
            # print >> sys.stderr, "r too small in assign_masses, m200b will be wrong!"
            # print >> sys.stderr, "edge_dens/dens_thresh[1] %.3f" % (cur_dens/dens_thresh[1])

    def max_halo_radius(self, int i):
        return max_halo_radius(&halos[i])

    def output_halos(self, np.int64_t idoffset, np.ndarray[np.float32_t, ndim=2] bbox):
        cdef float bounds[6]
        if idoffset is None: idoffset = 0
        if bbox is None:
            output_halos(idoffset, 0, 0, NULL) 
        else:
            for i in range(3):
                bounds[i] = bbox[i,0]
                bounds[i+3] = bbox[i,1]
            output_halos(idoffset, 0, 0, bounds) 

    def output_config(self):
        output_config(NULL) 

    def return_halos(self):
        cdef haloflat[:] haloview = <haloflat[:num_halos]> (<haloflat*> halos)
        return np.asarray(haloview)

    def finish(self):
        rockstar_cleanup()
        free_halos()

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def make_rockstar_fof(self, np.ndarray[np.int64_t, ndim=1] pind,
                                np.ndarray[np.int64_t, ndim=1] fof_tags,
                                np.ndarray[floating, ndim=2] pos,
                                np.ndarray[floating, ndim=2] vel):

        verbose = False
        # Define fof object

        # Find number of particles
        cdef np.int64_t i, j, k, ind
        global global_particles

        # Allocate space for correct number of particles
        cdef fof fof_obj

        cdef np.int64_t max_count = 0
        cdef np.int64_t next_tag, local_tag, last_fof_tag = -1
        fof_obj.num_p = 0
        j = 0
        # We're going to do one iteration to get the most frequent value.
        for i in range(pind.shape[0]):
            ind = pind[i]
            local_tag = fof_tags[ind]
            # Don't count the null group
            if local_tag == -1: continue
            if local_tag != last_fof_tag:
                if j > max_count:
                    max_count = j
                last_fof_tag = local_tag
                j = 1
            else:
                j += 1
        if j > max_count:
            max_count = j
        #print >> sys.stderr, "Most frequent occurrence: %s" % max_count
        fof_obj.particles = <particle*> malloc(max_count * sizeof(particle))
        j = 0
        cdef int counter = 0, ndone = 0
        cdef np.ndarray[np.int64_t, ndim=1] pcounts 
        pcounts = np.zeros(np.unique(fof_tags).size, dtype="int64")
        cdef np.int64_t frac = <np.int64_t> (pcounts.shape[0] / 20.0)
        free_halos()
        for i in range(pind.shape[0]):
            ind = pind[i]
            local_tag = fof_tags[ind]
            # Skip this one -- it means no group.
            if local_tag == -1:
                continue
            if i == pind.shape[0] - 1:
                next_tag = local_tag + 1
            else:
                next_tag = fof_tags[pind[i+1]]
            for k in range(3):
                fof_obj.particles[j].pos[k] = pos[ind,k]
                fof_obj.particles[j].pos[k+3] = vel[ind,k]
            fof_obj.particles[j].id = j
            fof_obj.num_p += 1
            j += 1
            # Now we check if we're the last one
            if local_tag != next_tag:
                pcounts[ndone] = fof_obj.num_p
                counter += 1
                ndone += 1
                if verbose and counter == frac:
                    print >> sys.stderr, "R*-ing % 5.1f%% done (%0.3f -> %0.3f)" % (
                        (100.0 * ndone)/pcounts.size,
                        fof_obj.particles[0].pos[2],
                        halos[num_halos - 1].pos[2])
                    counter = 0
                global_particles = &fof_obj.particles[0]
                find_subs(&fof_obj)
                # Now we reset
                fof_obj.num_p = j = 0
        free(fof_obj.particles)
        global_particles = NULL
        return pcounts



"""
Gadget frontend tests




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2015, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from collections import OrderedDict

from yt.testing import requires_file
from yt.utilities.answer_testing.framework import \
    data_dir_load, \
    requires_ds, \
    sph_answer
from yt.frontends.gadget.api import GadgetHDF5Dataset, GadgetDataset

isothermal_h5 = "IsothermalCollapse/snap_505.hdf5"
isothermal_bin = "IsothermalCollapse/snap_505"
BE_Gadget = "BigEndianGadgetBinary/BigEndianGadgetBinary"
LE_SnapFormat2 = "Gadget3-snap-format2/Gadget3-snap-format2"
keplerian_ring = "KeplerianRing/keplerian_ring_test_data.hdf5"

# This maps from field names to weight field names to use for projections
iso_fields = OrderedDict(
    [
        (("gas", "density"), None),
        (("gas", "temperature"), None),
        (("gas", "temperature"), ('gas', 'density')),
        (('gas', 'velocity_magnitude'), None),
        (("deposit", "all_density"), None),
        (("deposit", "all_count"), None),
        (("deposit", "all_cic"), None),
        (("deposit", "PartType0_density"), None),
    ]
)
iso_kwargs = dict(bounding_box=[[-3, 3], [-3, 3], [-3, 3]])


@requires_file(isothermal_h5)
@requires_file(isothermal_bin)
@requires_file(BE_Gadget)
@requires_file(LE_SnapFormat2)
def test_gadget_dataset():
    assert isinstance(data_dir_load(isothermal_h5, kwargs=iso_kwargs),
                      GadgetHDF5Dataset)
    assert isinstance(data_dir_load(isothermal_bin, kwargs=iso_kwargs),
                      GadgetDataset)
    assert isinstance(data_dir_load(BE_Gadget), GadgetDataset)
    assert isinstance(data_dir_load(LE_SnapFormat2), GadgetDataset)


@requires_file(keplerian_ring)
def test_non_cosmo_dataset():
    """
    Non-cosmological datasets may not have the cosmological parametrs in the
    Header. The code should fall back gracefully when they are not present,
    with the Redshift set to 0.
    """
    data = data_dir_load(keplerian_ring)
    assert data.current_redshift == 0.0
    assert data.cosmological_simulation == 0


@requires_ds(isothermal_h5)
def test_iso_collapse():
    ds = data_dir_load(isothermal_h5, kwargs=iso_kwargs)
    for test in sph_answer(ds, 'snap_505', 2**17, iso_fields):
        test_iso_collapse.__name__ = test.description
        yield test


@requires_ds(LE_SnapFormat2)
def test_pid_uniqueness():
    """
    ParticleIDs should be unique.
    """
    ds = data_dir_load(LE_SnapFormat2)
    ad = ds.all_data()
    pid = ad['ParticleIDs']
    assert len(pid) == len(set(pid.v))



from yt.analysis_modules.photon_simulator.api import \
    TableApecModel, XSpecThermalModel
from yt.testing import requires_module, fake_random_ds
from yt.utilities.answer_testing.framework import \
    GenericArrayTest
from yt.config import ytcfg

def setup():
    ytcfg["yt", "__withintesting"] = "True"

xray_data_dir = ytcfg.get("yt", "xray_data_dir")

ds = fake_random_ds(64)

@requires_module("xspec")
@requires_module("astropy")
def test_apec():

    settings = {"APECROOT":xray_data_dir+"/apec_v2.0.2"}
    xmod = XSpecThermalModel("apec", 0.1, 10.0, 10000, thermal_broad=True,
                             settings=settings)
    xmod.prepare_spectrum(0.2)

    xcspec, xmspec = xmod.get_spectrum(6.0)
    spec1 = xcspec+0.3*xmspec

    amod = TableApecModel(xray_data_dir, 0.1, 10.0, 
                          10000, thermal_broad=True)
    amod.prepare_spectrum(0.2)

    acspec, amspec = amod.get_spectrum(6.0)
    spec2 = acspec+0.3*amspec

    def spec1_test():
        return spec1.v
    def spec2_test():
        return spec2.v

    for test in [GenericArrayTest(ds, spec1_test),
                 GenericArrayTest(ds, spec2_test)]:
        test_apec.__name__ = test.description
        yield test

    xmod.cleanup_spectrum()



"""
Code to export from yt to RadMC3D



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
from yt.utilities.lib.write_array import \
    write_3D_array, write_3D_vector_array


class RadMC3DLayer:
    '''

    This class represents an AMR "layer" of the style described in
    the radmc3d manual. Unlike yt grids, layers may not have more
    than one parent, so level L grids will need to be split up
    if they straddle two or more level L - 1 grids.

    '''
    def __init__(self, level, parent, unique_id, LE, RE, dim):
        self.level = level
        self.parent = parent
        self.LeftEdge = LE
        self.RightEdge = RE
        self.ActiveDimensions = dim
        self.id = unique_id

    def get_overlap_with(self, grid):
        '''

        Returns the overlapping region between two Layers,
        or a layer and a grid. RE < LE means in any direction
        means no overlap.

        '''
        LE = np.maximum(self.LeftEdge,  grid.LeftEdge)
        RE = np.minimum(self.RightEdge, grid.RightEdge)
        return LE, RE

    def overlaps(self, grid):
        '''

        Returns whether or not this layer overlaps a given grid

        '''
        LE, RE = self.get_overlap_with(grid)
        if np.any(RE <= LE):
            return False
        else:
            return True


class RadMC3DWriter:
    '''

    This class provides a mechanism for writing out data files in a format
    readable by radmc3d. Currently, only the ASCII, "Layer" style file format
    is supported. For more information please see the radmc3d manual at:
    http://www.ita.uni-heidelberg.de/~dullemond/software/radmc-3d

    Parameters
    ----------

    ds : `Dataset`
        This is the dataset object corresponding to the
        simulation output to be written out.

    max_level : int
        An int corresponding to the maximum number of levels of refinement
        to include in the output. Often, this does not need to be very large
        as information on very high levels is frequently unobservable.
        Default = 2.

    Examples
    --------

    This will create a field called "DustDensity" and write it out to the
    file "dust_density.inp" in a form readable by RadMC3D.

    >>> import yt
    >>> from yt.analysis_modules.radmc3d_export.api import RadMC3DWriter

    >>> dust_to_gas = 0.01
    >>> def _DustDensity(field, data):
    ...     return dust_to_gas*data["Density"]
    >>> yt.add_field("DustDensity", function=_DustDensity)

    >>> ds = yt.load("galaxy0030/galaxy0030")

    >>> writer = RadMC3DWriter(ds)
    >>> writer.write_amr_grid()
    >>> writer.write_dust_file("DustDensity", "dust_density.inp")

    ---

    This example will create a field called "NumberDensityCO" and write it out
    to the file "numberdens_co.inp". It will also write out information about
    the gas velocity to "gas_velocity.inp" so that this broadening may be
    included in the radiative transfer calculation by radmc3d:

    >>> import yt
    >>> from yt.analysis_modules.radmc3d_export.api import RadMC3DWriter

    >>> x_co = 1.0e-4
    >>> mu_h = yt.Quantity(2.34e-24, 'g')
    >>> def _NumberDensityCO(field, data):
    ...     return (x_co/mu_h)*data["Density"]
    >>> yt.add_field("NumberDensityCO", function=_NumberDensityCO)

    >>> ds = yt.load("galaxy0030/galaxy0030")
    >>> writer = RadMC3DWriter(ds)

    >>> writer.write_amr_grid()
    >>> writer.write_line_file("NumberDensityCO", "numberdens_co.inp")
    >>> velocity_fields = ["velocity_x", "velocity_y", "velocity_z"]
    >>> writer.write_line_file(velocity_fields, "gas_velocity.inp")

    '''

    def __init__(self, ds, max_level=2):
        self.max_level = max_level
        self.cell_count = 0
        self.layers = []
        self.domain_dimensions = ds.domain_dimensions
        self.domain_left_edge = ds.domain_left_edge
        self.domain_right_edge = ds.domain_right_edge
        self.grid_filename = "amr_grid.inp"
        self.ds = ds

        base_layer = RadMC3DLayer(0, None, 0,
                                  self.domain_left_edge,
                                  self.domain_right_edge,
                                  self.domain_dimensions)

        self.layers.append(base_layer)
        self.cell_count += np.product(ds.domain_dimensions)

        sorted_grids = sorted(ds.index.grids, key=lambda x: x.Level)
        for grid in sorted_grids:
            if grid.Level <= self.max_level:
                self._add_grid_to_layers(grid)

    def _get_parents(self, grid):
        parents = []
        for potential_parent in self.layers:
            if potential_parent.level == grid.Level - 1:
                if potential_parent.overlaps(grid):
                    parents.append(potential_parent)
        return parents

    def _add_grid_to_layers(self, grid):
        parents = self._get_parents(grid)
        for parent in parents:
            LE, RE = parent.get_overlap_with(grid)
            N = (RE - LE) / grid.dds
            N = np.array([int(n + 0.5) for n in N])
            new_layer = RadMC3DLayer(grid.Level, parent.id,
                                     len(self.layers),
                                     LE, RE, N)
            self.layers.append(new_layer)
            self.cell_count += np.product(N)

    def write_amr_grid(self):
        '''
        This routine writes the "amr_grid.inp" file that describes the mesh
        radmc3d will use.

        '''
        dims = self.domain_dimensions
        LE = self.domain_left_edge
        RE = self.domain_right_edge

        # RadMC-3D wants the cell wall positions in cgs. Convert here:
        LE_cgs = LE.in_units('cm').d  # don't write the units, though
        RE_cgs = RE.in_units('cm').d

        # calculate cell wall positions
        xs = [str(x) for x in np.linspace(LE_cgs[0], RE_cgs[0], dims[0]+1)]
        ys = [str(y) for y in np.linspace(LE_cgs[1], RE_cgs[1], dims[1]+1)]
        zs = [str(z) for z in np.linspace(LE_cgs[2], RE_cgs[2], dims[2]+1)]

        # writer file header
        grid_file = open(self.grid_filename, 'w')
        grid_file.write('1 \n')  # iformat is always 1
        if self.max_level == 0:
            grid_file.write('0 \n')
        else:
            grid_file.write('10 \n')  # only layer-style files are supported
        grid_file.write('1 \n')  # only cartesian coordinates are supported
        grid_file.write('0 \n')
        grid_file.write('{}    {}    {} \n'.format(1, 1, 1))  # assume 3D
        grid_file.write('{}    {}    {} \n'.format(dims[0], dims[1], dims[2]))
        if self.max_level != 0:
            s = str(self.max_level) + '    ' + str(len(self.layers)-1) + '\n'
            grid_file.write(s)

        # write base grid cell wall positions
        for x in xs:
            grid_file.write(x + '    ')
        grid_file.write('\n')

        for y in ys:
            grid_file.write(y + '    ')
        grid_file.write('\n')

        for z in zs:
            grid_file.write(z + '    ')
        grid_file.write('\n')

        # write information about fine layers, skipping the base layer:
        for layer in self.layers[1:]:
            p = layer.parent
            dds = (layer.RightEdge - layer.LeftEdge) / (layer.ActiveDimensions)
            if p == 0:
                ind = (layer.LeftEdge - LE) / (2.0*dds) + 1
            else:
                parent_LE = np.zeros(3)
                for potential_parent in self.layers:
                    if potential_parent.id == p:
                        parent_LE = potential_parent.LeftEdge
                ind = (layer.LeftEdge - parent_LE) / (2.0*dds) + 1
            ix = int(ind[0]+0.5)
            iy = int(ind[1]+0.5)
            iz = int(ind[2]+0.5)
            nx, ny, nz = layer.ActiveDimensions / 2
            s = '{}    {}    {}    {}    {}    {}    {} \n'
            s = s.format(p, ix, iy, iz, nx, ny, nz)
            grid_file.write(s)

        grid_file.close()

    def _write_layer_data_to_file(self, fhandle, field, level, LE, dim):
        cg = self.ds.covering_grid(level, LE, dim, num_ghost_zones=1)
        if isinstance(field, list):
            data_x = cg[field[0]]
            data_y = cg[field[1]]
            data_z = cg[field[2]]
            write_3D_vector_array(data_x, data_y, data_z, fhandle)
        else:
            data = cg[field]
            write_3D_array(data, fhandle)

    def write_dust_file(self, field, filename):
        '''
        This method writes out fields in the format radmc3d needs to compute
        thermal dust emission. In particular, if you have a field called
        "DustDensity", you can write out a dust_density.inp file.

        Parameters
        ----------

        field : string
            The name of the field to be written out
        filename : string
            The name of the file to write the data to. The filenames radmc3d
            expects for its various modes of operations are described in the
            radmc3d manual.

        '''
        fhandle = open(filename, 'w')

        # write header
        fhandle.write('1 \n')
        fhandle.write(str(self.cell_count) + ' \n')
        fhandle.write('1 \n')

        # now write fine layers:
        for layer in self.layers:
            lev = layer.level
            if lev == 0:
                LE = self.domain_left_edge
                N = self.domain_dimensions
            else:
                LE = layer.LeftEdge
                N = layer.ActiveDimensions

            self._write_layer_data_to_file(fhandle, field, lev, LE, N)

        fhandle.close()

    def write_line_file(self, field, filename):
        '''
        This method writes out fields in the format radmc3d needs to compute
        line emission.

        Parameters
        ----------

        field : string or list of 3 strings
            If a string, the name of the field to be written out. If a list,
            three fields that will be written to the file as a vector quantity.
        filename : string
            The name of the file to write the data to. The filenames radmc3d
            expects for its various modes of operation are described in the
            radmc3d manual.

        '''
        fhandle = open(filename, 'w')

        # write header
        fhandle.write('1 \n')
        fhandle.write(str(self.cell_count) + ' \n')

        # now write fine layers:
        for layer in self.layers:
            lev = layer.level
            if lev == 0:
                LE = self.domain_left_edge
                N = self.domain_dimensions
            else:
                LE = layer.LeftEdge
                N = layer.ActiveDimensions

            self._write_layer_data_to_file(fhandle, field, lev, LE, N)

        fhandle.close()

    def write_source_files(self, sources, wavelengths):
        '''

        This function creates the stars.inp and wavelength_micron.inp
        files that RadMC3D uses for its dust continuum calculations.

        Parameters
        ----------

        sources: a list of RadMC3DSource objects
            A list that contains all the sources you would like yt
            to create
        wavelengths: np.array of float values
            An array listing the wavelength points you would like to
            use the radiative transfer calculation

        '''

        nstars = len(sources)
        nlam = len(wavelengths)

        filename = 'stars.inp'
        fhandle = open(filename, 'w')

        # write header
        fhandle.write('2 \n')  # a format flag that should always be 2
        fhandle.write('%d    %d \n' % (nstars, nlam))

        # write source information
        for source in sources:
            fhandle.write(str(source.radius) + ' ')
            fhandle.write(str(source.mass) + ' ')
            fhandle.write('%f %f %f' %(source.position[0], \
                                       source.position[1], \
                                       source.position[2]))
            fhandle.write('\n')

        # write wavelength informaton
        for wavelength in wavelengths:
            fhandle.write('%f \n' % wavelength)

        # finally write blackbody temperature for each source
        for source in sources:
            # the negative sign is a flag used internally
            # by RadMC3D to indicate that this is a blackbody
            # source
            fhandle.write('%f \n' % -source.temperature)

        # done with stars.inp
        fhandle.close()

        # now do the wavelength_micron.inp file
        filename = 'wavelength_micron.inp'
        fhandle = open(filename, 'w')

        fhandle.write('%d \n' % nlam)
        for wavelength in wavelengths:
            fhandle.write('%f \n' % wavelength)

        # done with both
        fhandle.close()


class RadMC3DSource:
    '''

    A class that contains the data associated with a single RadMC3D photon source.
    This is designed to help export data about the stars in a dataset into a format 
    that can be read in by RadMC3D. Although RadMC3D can handle non-blackbody 
    sources, here we assume that the source is a blackbody with a given temperature.

    Parameters
    ----------

    radius: float
        The size of the source in cm
    mass: float 
        The mass of the source in g
    position: list of floats
        The x, y, and z coordinates of the source, in cm
    temperature: float
        The blackbody temperature of the source, in K

    '''

    def __init__(self, radius, mass, position, temperature):
        self.radius = radius
        self.mass = mass
        self.position = position
        self.temperature = temperature

        # some basic sanity checks
        assert(self.radius > 0.0)
        assert(self.mass > 0.0)
        assert(self.temperature > 0)
        assert(len(self.position) == 3)  # 3D only, please



"""
API for two_point_functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------


from yt.funcs import issue_deprecation_warning

issue_deprecation_warning(
    "The two_point_functions module has been deprecated. This code has been "
    "moved to the yt attic (https://github.com/yt-project/yt_attic) and "
    "will be removed in a future release.")

from .two_point_functions import \
    TwoPointFunctions, \
    FcnSet



"""MOAB-specific fields


"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
from yt.funcs import mylog
from yt.utilities.io_handler import BaseIOHandler


def field_dname(field_name):
    return "/tstt/elements/Hex8/tags/{0}".format(field_name)

# TODO all particle bits were removed
class IOHandlerMoabH5MHex8(BaseIOHandler):
    _dataset_type = "moab_hex8"

    def __init__(self, ds):
        super(IOHandlerMoabH5MHex8, self).__init__(ds)
        self._handle = ds._handle

    def _read_fluid_selection(self, chunks, selector, fields, size):
        chunks = list(chunks)
        assert(len(chunks) == 1)
        fhandle = self._handle
        rv = {}
        for field in fields:
            ftype, fname = field
            rv[field] = np.empty(size, dtype=fhandle[field_dname(fname)].dtype)
        ngrids = sum(len(chunk.objs) for chunk in chunks)
        mylog.debug("Reading %s cells of %s fields in %s blocks",
                    size, [fname for ft, fn in fields], ngrids)
        for field in fields:
            ftype, fname = field
            ds = np.array(fhandle[field_dname(fname)][:], dtype="float64")
            ind = 0
            for chunk in chunks:
                for g in chunk.objs:
                    ind += g.select(selector, ds, rv[field], ind) # caches
        return rv

class IOHandlerMoabPyneHex8(BaseIOHandler):
    _dataset_type = "moab_hex8_pyne"

    def _read_fluid_selection(self, chunks, selector, fields, size):
        chunks = list(chunks)
        assert(len(chunks) == 1)
        rv = {}
        pyne_mesh = self.ds.pyne_mesh
        for field in fields:
            rv[field] = np.empty(size, dtype="float64")
        ngrids = sum(len(chunk.objs) for chunk in chunks)
        mylog.debug("Reading %s cells of %s fields in %s blocks",
                    size, [fname for ftype, fname in fields], ngrids)
        for field in fields:
            ftype, fname = field
            if pyne_mesh.structured:
                tag = pyne_mesh.mesh.getTagHandle('idx')
                indices = [tag[ent] for ent in pyne_mesh.structured_iterate_hex()]
            else:
                indices = slice(None)
            ds = np.asarray(getattr(pyne_mesh, fname)[indices], 'float64')

            ind = 0
            for chunk in chunks:
                for g in chunk.objs:
                    ind += g.select(selector, ds, rv[field], ind) # caches
        return rv



"""
GAMER-specific fields



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.fields.field_info_container import FieldInfoContainer
from yt.utilities.physical_constants import mh, boltzmann_constant_cgs

b_units   = "code_magnetic"
pre_units = "code_mass / (code_length*code_time**2)"
erg_units = "code_mass / (code_length*code_time**2)"
rho_units = "code_mass / code_length**3"
mom_units = "code_mass / (code_length**2*code_time)"
vel_units = "code_velocity"
pot_units = "code_length**2/code_time**2"

psi_units = "code_mass**(1/2) / code_length**(3/2)"


class GAMERFieldInfo(FieldInfoContainer):
    known_other_fields = (
        # hydro fields on disk (GAMER outputs conservative variables)
        ( "Dens",     (rho_units, ["density"],                  r"\rho") ),
        ( "MomX",     (mom_units, ["momentum_x"],               None   ) ),
        ( "MomY",     (mom_units, ["momentum_y"],               None   ) ),
        ( "MomZ",     (mom_units, ["momentum_z"],               None   ) ),
        ( "Engy",     (erg_units, ["total_energy_per_volume"],  None   ) ),
        ( "Pote",     (pot_units, ["gravitational_potential"],  None   ) ),

        # psiDM fields on disk
        ( "Real",     (psi_units, ["psidm_real_part"],          None   ) ),
        ( "Imag",     (psi_units, ["psidm_imaginary_part"],     None   ) ),

        # particle fields on disk (deposited onto grids)
        ("ParDens",   (rho_units, ["particle_density_on_grid"], None   ) ),
        ("TotalDens", (rho_units, ["total_density_on_grid"],    None   ) ),
    )

    known_particle_fields = (
        ( "ParMass", ("code_mass",     ["particle_mass"],       None) ),
        ( "ParPosX", ("code_length",   ["particle_position_x"], None) ),
        ( "ParPosY", ("code_length",   ["particle_position_y"], None) ),
        ( "ParPosZ", ("code_length",   ["particle_position_z"], None) ),
        ( "ParVelX", ("code_velocity", ["particle_velocity_x"], None) ),
        ( "ParVelY", ("code_velocity", ["particle_velocity_y"], None) ),
        ( "ParVelZ", ("code_velocity", ["particle_velocity_z"], None) ),
    )

    def __init__(self, ds, field_list):
        super(GAMERFieldInfo, self).__init__(ds, field_list)

    # add primitive and other derived variables
    def setup_fluid_fields(self):
        unit_system = self.ds.unit_system

        # velocity
        def velocity_xyz(v):
            def _velocity(field, data):
                return data["gas", "momentum_%s"%v] / data["gas","density"]
            return _velocity
        for v in "xyz":
            self.add_field( ("gas","velocity_%s"%v), sampling_type="cell",
                            function = velocity_xyz(v),
                            units = unit_system["velocity"] )

        # ============================================================================
        # note that yt internal fields assume
        #    [thermal_energy]          = [energy per mass]
        #    [kinetic_energy]          = [energy per volume]
        # and we further adopt
        #    [total_energy]            = [energy per mass]
        #    [total_energy_per_volume] = [energy per volume]
        # ============================================================================

        # kinetic energy per volume
        def ek(data):
            return 0.5*( data["gamer","MomX"]**2 +
                         data["gamer","MomY"]**2 +
                         data["gamer","MomZ"]**2 ) / data["gamer","Dens"]

        # thermal energy per volume
        def et(data):
            return data["gamer","Engy"] - ek(data)

        # thermal energy per mass (i.e., specific)
        def _thermal_energy(field, data):
            return et(data) / data["gamer","Dens"]
        self.add_field( ("gas","thermal_energy"), sampling_type="cell",
                        function = _thermal_energy,
                        units = unit_system["specific_energy"] )

        # total energy per mass
        def _total_energy(field, data):
            return data["gamer","Engy"] / data["gamer","Dens"]
        self.add_field( ("gas","total_energy"), sampling_type="cell",
                        function = _total_energy,
                        units = unit_system["specific_energy"] )

        # pressure
        def _pressure(field, data):
            return et(data)*(data.ds.gamma-1.0)
        self.add_field( ("gas","pressure"), sampling_type="cell",
                        function = _pressure,
                        units = unit_system["pressure"] )

        # temperature
        def _temperature(field, data):
            return data.ds.mu*mh*data["gas","pressure"] / \
                   (data["gas","density"]*boltzmann_constant_cgs)
        self.add_field( ("gas","temperature"), sampling_type="cell",
                        function = _temperature,
                        units = unit_system["temperature"] )

    def setup_particle_fields(self, ptype):
        super(GAMERFieldInfo, self).setup_particle_fields(ptype)



"""
Tests for AMRSlice


"""

# ----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# ----------------------------------------------------------------------------
import os
import numpy as np
import tempfile
from yt.testing import \
    fake_random_ds, assert_equal
from yt.units.unit_object import Unit


def setup():
    from yt.config import ytcfg
    ytcfg["yt", "__withintesting"] = "True"


def teardown_func(fns):
    for fn in fns:
        try:
            os.remove(fn)
        except OSError:
            pass


def test_slice():
    fns = []
    grid_eps = np.finfo(np.float64).eps
    for nprocs in [8, 1]:
        # We want to test both 1 proc and 8 procs, to make sure that
        # parallelism isn't broken
        ds = fake_random_ds(64, nprocs=nprocs)
        dims = ds.domain_dimensions
        xn, yn, zn = ds.domain_dimensions
        dx = ds.arr(1.0 / (ds.domain_dimensions * 2), 'code_length')
        xi, yi, zi = ds.domain_left_edge + dx
        xf, yf, zf = ds.domain_right_edge - dx
        coords = np.mgrid[xi:xf:xn * 1j, yi:yf:yn * 1j, zi:zf:zn * 1j]
        uc = [np.unique(c) for c in coords]
        slc_pos = 0.5
        # Some simple slice tests with single grids
        for ax, an in enumerate("xyz"):
            xax = ds.coordinates.x_axis[ax]
            yax = ds.coordinates.y_axis[ax]
            for wf in ["density", None]:
                slc = ds.slice(ax, slc_pos)
                shifted_slc = ds.slice(ax, slc_pos + grid_eps)
                assert_equal(slc["ones"].sum(), slc["ones"].size)
                assert_equal(slc["ones"].min(), 1.0)
                assert_equal(slc["ones"].max(), 1.0)
                assert_equal(np.unique(slc["px"]), uc[xax])
                assert_equal(np.unique(slc["py"]), uc[yax])
                assert_equal(np.unique(slc["pdx"]), 0.5 / dims[xax])
                assert_equal(np.unique(slc["pdy"]), 0.5 / dims[yax])
                pw = slc.to_pw(fields='density')
                for p in pw.plots.values():
                    tmpfd, tmpname = tempfile.mkstemp(suffix='.png')
                    os.close(tmpfd)
                    p.save(name=tmpname)
                    fns.append(tmpname)
                for width in [(1.0, 'unitary'), 1.0, ds.quan(0.5, 'code_length')]:
                    frb = slc.to_frb((1.0, 'unitary'), 64)
                    shifted_frb = shifted_slc.to_frb((1.0, 'unitary'), 64)
                    for slc_field in ['ones', 'density']:
                        fi = ds._get_field_info(slc_field)
                        assert_equal(frb[slc_field].info['data_source'],
                                     slc.__str__())
                        assert_equal(frb[slc_field].info['axis'], ax)
                        assert_equal(frb[slc_field].info['field'], slc_field)
                        assert_equal(frb[slc_field].units, Unit(fi.units))
                        assert_equal(frb[slc_field].info['xlim'],
                                     frb.bounds[:2])
                        assert_equal(frb[slc_field].info['ylim'],
                                     frb.bounds[2:])
                        assert_equal(frb[slc_field].info['center'], slc.center)
                        assert_equal(frb[slc_field].info['coord'], slc_pos)
                        assert_equal(frb[slc_field], shifted_frb[slc_field])
            assert_equal(wf, None)
    teardown_func(fns)


def test_slice_over_edges():
    ds = fake_random_ds(64, nprocs=8, fields=["density"], negative=[False])
    slc = ds.slice(0, 0.0)
    slc["density"]
    slc = ds.slice(1, 0.5)
    slc["density"]


def test_slice_over_outer_boundary():
    ds = fake_random_ds(64, nprocs=8, fields=["density"], negative=[False])
    slc = ds.slice(2, 1.0)
    slc["density"]
    assert_equal(slc["density"].size, 0)



"""
Fields specific to Enzo-P



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.fields.field_info_container import \
    FieldInfoContainer

rho_units = "code_mass / code_length**3"
vel_units = "code_velocity"
acc_units = "code_velocity / code_time"
energy_units = "code_velocity**2"

known_species_names = {
}

class EnzoPFieldInfo(FieldInfoContainer):
    known_other_fields = (
        ("velocity_x", (vel_units, ["velocity_x"], None)),
        ("velocity_y", (vel_units, ["velocity_y"], None)),
        ("velocity_z", (vel_units, ["velocity_z"], None)),
        ("acceleration_x", (acc_units, ["acceleration_x"], None)),
        ("acceleration_y", (acc_units, ["acceleration_y"], None)),
        ("acceleration_z", (acc_units, ["acceleration_z"], None)),
        ("density", (rho_units, ["density"], None)),
        ("total_density", (rho_units, ["total_density"], None)),
        ("total_energy", (energy_units, ["total_energy"], None)),
        ("internal_energy", (energy_units, ["internal_energy"], None)),
    )

    known_particle_fields = (
        ("x", ("code_length", [], None)),
        ("y", ("code_length", [], None)),
        ("z", ("code_length", [], None)),
        ("vx", (vel_units, [], None)),
        ("vy", (vel_units, [], None)),
        ("vz", (vel_units, [], None)),
        ("ax", (acc_units, [], None)),
        ("ay", (acc_units, [], None)),
        ("az", (acc_units, [], None)),
        ("mass", ("code_mass", [], None)),
    )
